/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CBaseMailConnect.h  -  description
                             -------------------
    begin                : 25 June 2003
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/
#ifndef __CBASEMAILCONNECT_H__
#define __CBASEMAILCONNECT_H__

#include <sptk3/sptk.h>
#include <sptk3/CBuffer.h>
#include <sptk3/CMailMessageBody.h>

#include <string>

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

/// @brief Base mail socket
///
/// CBaseMailConnect class is the base class for mail message components
class CBaseMailConnect {
protected:
    std::string      m_from;     ///< Mail FROM: a single e-mail address in format: "Jonh Doe <jonhd@noname.com>"
    std::string      m_to;       ///< Mail TO: semicolon-separated string of addresses in format: "Jonh Doe <jonhd@noname.com>; Jane Doe <janed@noname.com>"
    std::string      m_cc;       ///< Mail CC: semicolon-separated string of addresses in format: "Jonh Doe <jonhd@noname.com>; Jane Doe <janed@noname.com>"
    std::string      m_bcc;      ///< Mail BCC: semicolon-separated string of addresses in format: "Jonh Doe <jonhd@noname.com>; Jane Doe <janed@noname.com>"
    std::string      m_subject;  ///< Mail SUBJECT:
    CMailMessageBody m_body;     ///< Mail text (plain-text and html parts of the message)
    std::string      m_attachments; ///< The list of attachment files separated with ';'

    CBuffer m_messageBuffer; ///< Internal message buffer

    /// Encoding the message into internal message buffer
    void mimeFile(std::string fileName,std::string fileAlias,CBuffer& buffer);

public:
    /// Default constructor
    CBaseMailConnect() {}

    /// Default destructor
    virtual ~CBaseMailConnect() {}

    /// Method from() returns the current value of 'FROM:' field
    /// of e-mail message.
    /// @returns a single e-mail address.
    std::string from() const        {
        return m_from;
    }

    /// Method from() sets the current value of 'FROM:' field
    /// of e-mail message.
    /// @param addr should be an e-mail address in format:
    /// Real sender name <sender@host.net>. The example: John Doe <johnd@unknown.org>
    void from(std::string addr)     {
        m_from = addr;
    }

    /// Method to() returns the current value of 'TO:' field
    /// of e-mail message
    std::string to() const          {
        return m_to;
    }

    /// Method from() sets the current value of 'TO:' field
    /// of e-mail message.
    /// @param addr should be a semicolon-separated list of one or more e-mail addresses in format:
    /// Real sender name <sender@host.net>. The example: John Doe <johnd@unknown.org>
    void to(std::string addr)       {
        m_to = addr;
    }

    /// Method cc() returns the current value of 'CC:' field
    /// of e-mail message
    /// @returns a list of e-mail addresses. See method to() description for format
    std::string cc() const          {
        return m_cc;
    }

    /// Method cc() sets the current value of 'CC:' field
    /// of e-mail message.
    /// @param addr should be a semicolon-separated list of one or more e-mail addresses in format:
    /// Real sender name <sender@host.net>. The example: John Doe <johnd@unknown.org>
    void cc(std::string addr)       {
        m_cc = addr;
    }

    /// Method bcc() returns the current value of 'BCC:' field
    /// of e-mail message.
    /// @returns a list of e-mail addresses. See method to() description for format
    std::string bcc() const         {
        return m_bcc;
    }

    /// Method bcc() sets the current value of 'BCC:' field
    /// of e-mail message.
    /// @param addr should be a semicolon-separated list of one or more e-mail addresses in format:
    /// Real sender name <sender@host.net>. The example: John Doe <johnd@unknown.org>
    void bcc(std::string addr)      {
        m_bcc = addr;
    }

    /// Method subject() returns the current value of 'SUBJECT:' field
    /// of e-mail message.
    /// @returns current message subject
    std::string subject() const     {
        return m_subject;
    }

    /// Method subject() sets the current value of 'BCC:' field
    /// of e-mail message.
    /// @param subj A message subject
    void subject(std::string subj)  {
        m_subject = subj;
    }

    /// Method subject() returns the current plain text part
    /// of e-mail message.
    /// @returns current message plain-text part
    std::string body() const        {
        return m_body.text();
    }

    /// @brief Sets the current plain text part of e-mail message.
    ///
    /// @param body std::string&, message body
    /// @param smtp bool, do we need special pre-processing for SMTP?
    void body(const std::string& body,bool smtp)     {
        m_body.text(body,smtp);
    }

    /// Method attachments() returns the current semicolon-separated
    /// list of attachments of e-mail message. Example: "readme.txt;readme.doc".
    /// @returns current message list of attachments
    std::string attachments() const {
        return m_attachments;
    }

    /// Method attachments() sets the current semicolon-separated
    /// list of attachments of e-mail message. Example: "readme.txt;readme.doc".
    /// @param attachments current message list of attachments
    void attachments(std::string attachments) {
        m_attachments = attachments;
    }

    /// Method messageBuffer() returns the reference to the internal current message text completely
    /// prepared for sending, as described in RFC-822 message format. It only makes sense to use it after call to sendMessage().
    /// @returns reference to current message text
    const CBuffer& messageBuffer() const {
        return m_messageBuffer;
    }

    /// Method mimeMessage() encodes the message components into RFC-822 message format.
    /// @param buffer A buffer to put the encoded RFC-822 format message
    void mimeMessage(CBuffer& buffer);

    /// Method sendMessage() builds an RFC-822 format message out of message parameters,
    /// and sends it. Should be implemented in derived classes.
    virtual void sendMessage() = 0;
};
/// @}
}
#endif
