/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CDataSource.h  -  description
                             -------------------
    begin                : Jun 20 2003
    copyright            : (C) 1999-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CDATASOURCE_H__
#define __CDATASOURCE_H__

#include <sptk3/CVariant.h>
#include <sptk3/CFieldList.h>
#include <sptk3/cxml>

class Fl_Group;

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

/// @brief Universal data source for many SPTK widgets.
///
/// It's designed as a base class for multiple datasources available in SPTK.
/// The main idea is to provide the simple interface that allows to open the datasource
/// with certain parameters and read or write the datasource fields. And don't you forget to close it :)
class CDataSource {
    friend class Fl_Group;

protected:
    /// Loads datasource data. Should be implemented in derived class
    virtual bool loadData() {
        return true;
    }
    /// Saves data from datasource. Should be implemented in derived class
    virtual bool saveData() {
        return true;
    }

public:
    /// Default constructor
    CDataSource() {}

    /// Destructor
    virtual ~CDataSource() {}

public:
    /// Field access by the field index, const version.
    /// Purely virtual. Should be implemented in derived class
    /// @param fieldIndex uint32_t, field index
    /// @returns field reference
    virtual const CField& operator [] (uint32_t fieldIndex) const = 0;

    /// Field access by the field index, non-const version
    /// Purely virtual. Should be implemented in derived class
    /// @param fieldIndex uint32_t, field index
    /// @returns field reference
    virtual CField&       operator [] (uint32_t fieldIndex) = 0;

    /// Field access by the field name, const version.
    /// Purely virtual. Should be implemented in derived class
    /// @param fieldName const char *, field name
    /// @returns field reference
    virtual const CField& operator [] (const char *fieldName) const = 0;

    /// Field access by the field name, const version.
    /// Purely virtual. Should be implemented in derived class
    /// @param fieldName const char *, field name
    /// @returns field reference
    virtual CField&       operator [] (const char *fieldName) = 0;

    /// Returns field count in the datasource.
    /// Purely virtual. Should be implemented in derived class
    /// @returns field count
    virtual uint32_t        fieldCount() const = 0;
public:

    /// Returns record count in the datasource.
    /// Purely virtual. Should be implemented in derived class
    /// @returns record count
    virtual uint32_t        recordCount() const = 0;

    /// Returns user_data associated with the datasource.
    virtual void         *user_data() const {
        return 0L;
    }
public:
    /// Reads the field by name from the datasource.
    /// Purely virtual. Should be implemented in derived class
    /// @param fieldName const char *, field name
    /// @param fieldValue CVariant, field value
    virtual bool          readField(const char *fieldName,CVariant& fieldValue) = 0;

    /// Writes the field by name from the datasource.
    /// Purely virtual. Should be implemented in derived class
    /// @param fieldName const char *, field name
    /// @param fieldValue CVariant, field value
    virtual bool          writeField(const char *fieldName,const CVariant& fieldValue) = 0;
public:

    /// Opens the datasource. Implemented in derved class.
    virtual bool          open()  {
        return false;
    }

    /// Closes the datasource. Implemented in derved class.
    virtual bool          close() {
        return false;
    }

    /// Moves to the first record of the datasource. Implemented in derved class.
    virtual bool          first() {
        return false;
    }

    /// Moves to the next record of the datasource. Implemented in derved class.
    virtual bool          next()  {
        return false;
    }

    /// Moves to the prior record of the datasource. Implemented in derved class.
    virtual bool          prior() {
        return false;
    }

    /// Moves to the last record of the datasource. Implemented in derved class.
    virtual bool          last()  {
        return false;
    }

    /// Moves to the specified record position of the datasource. Implemented in derved class.
    virtual bool          find(CVariant position) {
        return false;
    }

    /// Returns true if there are no more records in the datasource. Implemented in derved class.
    virtual bool          eof() const  {
        return false;
    }

    /// Loads data into the datasource
    bool load();

    /// Saves data from the datasource
    bool save();

    /// @brief Saves dataset row data into XML
    ///
    /// If the compactXmlMode is true, the node would have fields presented as attributues.
    /// Otherwise, the fields are stored as subnodes.
    /// @param node CXmlNode&, the XML node to fill in
    /// @param compactXmlMode bool, compact XML flag
    void rowToXML(CXmlNode& node,bool compactXmlMode) const;

    /// @brief Saves data into XML
    ///
    /// Opens the dataset, reads every row, and closes dataset.
    /// For every row in the dataset, creates the node with the name nodeName.
    /// If the compactXmlMode is true, the nodes would have fields presented as attributues.
    /// Otherwise, the fields are stored as subnodes.
    /// @param parentNode CXmlNode&, the XML node to add subnodes to
    /// @param nodeName std::string, the name for subnodes
    /// @param compactXmlMode bool, compact XML flag
    virtual void toXML(CXmlNode& parentNode,std::string nodeName,bool compactXmlMode);
};

/// @}
}
#endif
