/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CDateTime.h  -  description
                             -------------------
    begin                : Tue Dec 14 1999
    copyright            : (C) 1999-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CDATETIME_H__
#define __CDATETIME_H__

#include <sptk3/sptk.h>
#include <string>

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

class CDateTime;
class CDateTimeFormat;

/// @brief Date and Time value.
///
/// Represents the date and time value. This value is stored as
/// a floating point number. Allows to synchronize the Now() time
/// with the external date/time, without affecting the local host
/// system time.
class SP_EXPORT CDateTime {
    friend class CDateTimeFormat;
protected:

    /// Internal decode date operation into y,m,d
    static void   decodeDate(const double dt,int16_t& y,int16_t& m,int16_t& d);

    /// Internal decode time operation into h,m,s,ms
    static void   decodeTime(const double dt,int16_t& h,int16_t& m,int16_t& s,int16_t& ms);

    /// Internal encode date operation from y,m,d
    static void   encodeDate(double &dt,int16_t y=0,int16_t m=0,int16_t d=0);

    /// Internal encode date operation from string
    static void   encodeDate(double &dt,const char *dat);

    /// Internal encode timee operation from h,m,s,ms
    static void   encodeTime(double &dt,int16_t h=0,int16_t m=0,int16_t s=0,int16_t ms=0);

    /// Internal encode timee operation from string
    static void   encodeTime(double &dt,const char *tim);

    /// Returns true for the leap year
    static int    isLeapYear(const int16_t year) {
        return (year&3) == 0 && year%100 || year%400 == 0;
    }

protected:

    double                m_dateTime;             ///< Actual date and time value

    /// The offset from current' system time for synchronization
    /// with outside system
    static double         m_dateTimeOffset;
    static bool           m_time24Mode;           ///< System's time mode
public:

    static char           dateFormat[32];         ///< System's date format
    static char           fullTimeFormat[32];     ///< System's time format
    static char           shortTimeFormat[32];    ///< System's time format
    static char           datePartsOrder[4];      ///< System's date parts order
    static char           dateSeparator;          ///< System's date separator
    static char           timeSeparator;          ///< System's time separator
    static std::string    weekDayNames[7];        ///< The locale-defined weekday names
    static std::string    monthNames[12];         ///< The locale-defined weekday names
    static std::string    timeZoneName;           ///< Time zone abbbreviastion
    static int            timeZoneOffset;         ///< Time zone offset from GMT in hours

public:

    /// @brief Constructor
    /// @param y int16_t, year
    /// @param m int16_t, month
    /// @param d int16_t, day
    /// @param h int16_t, hour
    /// @param mm int16_t, minute
    /// @param s int16_t, second
    CDateTime (int16_t y,int16_t m,int16_t d,int16_t h=0,int16_t mm=0,int16_t s=0);

    /// @brief Constructor
    /// @param dateStr const char *, date string
    CDateTime (const char * dateStr);

    /// @brief Copy constructor
    CDateTime (const CDateTime &dt);

    /// @brief Constructor
    /// @param dt double, floating point date and time value
    CDateTime (const double dt=0);

    /// @brief Conversion to double
    operator double () const;

    /// @brief Assignment
    void operator = (const CDateTime& date);

    /// @brief Assignment
    void operator = (const char * dat);

    /// @brief Addition, another CDateTime
    CDateTime  operator + (CDateTime& dt);

    /// @brief Substruction, another CDateTime
    CDateTime  operator - (CDateTime& dt);

    /// @brief Increment by another CDateTime
    CDateTime& operator += (CDateTime& dt);

    /// @brief Decrement by another CDateTime
    CDateTime& operator -= (CDateTime& dt);

    /// @brief Increment by day, prefix
    CDateTime& operator ++ ();

    /// @brief Increment by day, postfix
    CDateTime& operator ++ (int);

    /// @brief Decrement by day, prefix
    CDateTime& operator -- ();

    /// @brief Decrement by day, postfix
    CDateTime& operator -- (int);

    /// @brief Print the date into str
    void formatDate(char *str) const;

    /// @brief Print the date into str
    void formatTime(char *str,bool ampm=true,bool showSeconds=false) const;

    /// @brief Set the current date and time for this program only.
    ///
    /// The system time is not affected. Useful for synchronization between
    /// different hosts' programs.
    static void Now(CDateTime dt);

    /// @brief Reports the system date and time.
    static CDateTime System();

    /// @brief Reports the current date and time.
    static CDateTime Now();

    /// @brief Reports the current date.
    static CDateTime Date();

    /// @brief Reports the current time.
    static CDateTime Time();

    /// @brief Reports the current time of day in milliseconds
    ///
    /// This is fast method to get a time of day without considering dateTimeOffset.
    /// It's useful for measuring short time intervals (shorter than one day, anyway).
    static uint32_t TimeOfDayMs();

    /// @brief Converts C time into CDateTime
    static CDateTime convertCTime(const time_t tt);

    /// @brief Reports the number of days in the month in this date (1..31)
    int16_t daysInMonth() const;

    /// @brief Reports the day of the week in this date (1..7)
    int16_t dayOfWeek(void) const;

    /// @brief Reports the day since the beginning of the year in this date
    int16_t dayOfYear()  const;

    /// @brief Reports the day of the week name in this date ('Sunday'..'Saturday')
    std::string dayOfWeekName(void) const;

    /// @brief Reports the month name in this date ('Sunday'..'Saturday')
    std::string monthName() const;

    /// @brief Reports the date as an integer
    uint32_t date() const;

    /// @brief Reports the day of month (1..31)
    int16_t day() const;

    /// @brief Reports the month number (1..12)
    int16_t month() const;

    /// @brief Reports the year
    int16_t year() const;
public:

    /// @brief Returns date as a string
    std::string dateString() const;

    /// @brief Returns time as a string
    std::string timeString(bool showSeconds=false) const;

    /// @brief Returns date and time as a string
    operator std::string () const {
        return dateString() + " " + timeString();
    }
public:

    /// @brief Decodes date into y,m,d
    void decodeDate(int16_t *y,int16_t *m,int16_t *d) const {
        decodeDate(m_dateTime,*y,*m,*d);
    }

    /// @brief Decodes time into h,m,s,ms
    void decodeTime(int16_t *h,int16_t *m,int16_t *s,int16_t *ms) const {
        decodeTime(m_dateTime,*h,*m,*s,*ms);
    }

    /// @brief Returns system's time mode.
    static bool time24Mode() {
        return m_time24Mode;
    }

    /// @brief Sets system's time mode
    static void time24Mode(bool t24mode);
};
/// @}
}

/// @brief Compares CDatetime values
bool operator <  (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Compares CDateTime values
bool operator <= (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Compares CDateTime values
bool operator >  (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Compares CDatetime values
bool operator >= (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Compares CDatetime values
bool operator == (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Compares CDatetime values
bool operator != (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Adds two CDatetime values
sptk::CDateTime operator + (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

/// @brief Subtracts two CDatetime values
sptk::CDateTime operator - (const sptk::CDateTime &dt1, const sptk::CDateTime &dt2);

#endif
