/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CMemoryDS.h  -  description
                             -------------------
    begin                : Mon Apr 17 2000
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CMEMORYDS_H__
#define __CMEMORYDS_H__

#include <sptk3/sptk.h>
#include <sptk3/CFieldList.h>
#include <sptk3/CDataSource.h>
#include <sptk3/CException.h>
#include <vector>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Base (memory) datasource
///
/// Class CMemoryDS implements a special case of the datasource when the data
/// can be loaded all at once, in the datasource open() operation. It's a base class
/// for several actual datasources.
class CMemoryDS : public CDataSource  {
protected:
    std::vector<CFieldList *>  m_list;            ///< Internal list of the dataset records
    CFieldList*                m_current;         ///< Current record in the dataset.
    uint32_t                   m_currentIndex;    ///< The index of the current record.
    bool                       m_eof;             ///< EOF flag for sequentual reading first(),next()..next().

protected:

    /// Default constructor is protected, to prevent creating of the instance of that class
    CMemoryDS() : CDataSource(), m_current(0L), m_currentIndex(0) { }

public:

    /// Destructor
    virtual ~CMemoryDS() {
        close();
    }

    /// Clears all the records
    virtual void clear();

    /// Field access by the field index, const version.
    /// @param fieldIndex int, field index
    /// @returns field reference
    virtual const CField& operator [] (uint32_t fieldIndex) const;

    /// Field access by the field index, non-const version.
    /// @param fieldIndex int, field index
    /// @returns field reference
    virtual CField&       operator [] (uint32_t fieldIndex);

    /// Field access by the field name, const version.
    /// @param fieldName const char *, field name
    /// @returns field reference
    virtual const CField& operator [] (const char *fieldName) const;

    /// Field access by the field name, non-const version.
    /// @param fieldName const char *, field name
    /// @returns field reference
    virtual CField&       operator [] (const char *fieldName);

    /// Returns user_data associated with the datasource.
    virtual void             *user_data() const {
        return m_current->user_data();
    }

    /// Returns field count in the datasource.
    /// @returns field count
    virtual uint32_t           fieldCount() const;

    /// Returns record count in the datasource.
    /// @returns record count
    virtual uint32_t           recordCount() const;

    /// Reads the field by name from the datasource.
    /// @param fieldName const char *, field name
    /// @param fieldValue CVariant, field value
    virtual bool              readField(const char *fieldName,CVariant& fieldValue);

    /// Writes the field by name from the datasource.
    /// @param fieldName const char *, field name
    /// @param fieldValue CVariant, field value
    virtual bool              writeField(const char *fieldName,const CVariant& fieldValue);

    /// Opens the datasource. Implemented in derved class.
    virtual bool              open() {
        throw CException("Not implemented yet");
    }

    /// Closes the datasource.
    virtual bool              close();

    /// Moves to the first record of the datasource.
    virtual bool              first();

    /// Moves to the next record of the datasource.
    virtual bool              next();

    /// Moves to the prior record of the datasource.
    virtual bool              prior();

    /// Moves to the last record of the datasource.
    virtual bool              last();

    /// Finds the record by the record position (defined by record's user_data or key).
    virtual bool              find(CVariant position);

    /// Returns true if there are no more records in the datasource. Implemented in derved class.
    virtual bool              eof() const {
        return m_eof;
    }
};
/// @}
}
#endif
