/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CRegistry.h  -  description
                             -------------------
    begin                : Tue Mar 26 2002
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CREGISTRY_H__
#define __CREGISTRY_H__

#include <sptk3/sptk.h>
#include <sptk3/CStrings.h>
#include <sptk3/istring.h>
#include <sptk3/cxml>

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

class CRegistry;

/// @brief Registry modes
///
/// Modes for the registry. User registry is stored in home directory
/// (if home directory is avalable). Program registry is stored in the program's
/// startup directory.
enum CRegistryMode { USER_REGISTRY=1, PROGRAM_REGISTRY=2 };


/// @brief Registry
///
/// Works with INI and XML configuration-files.
/// Class allows to read both INI and XML files and write the output file
/// in the same format, or change the output format before saving the file.
class SP_EXPORT CRegistry : public CXmlDoc {
    std::string         m_fileName;            ///< The registry file name

    /// @brief Prepares program registry directory
    void prepareDirectory();

    /// @brief Saves a node into CStrings
    ///
    /// The node is saved in INI file format.
    /// @param outputData CStrings&, the string list to save data into
    /// @param node CXmlNode*, the XML node to save
    /// @param currentPath string, current path to the parent node
    void save(CStrings& outputData,CXmlNode* node,std::string currentPath);

    /// @brief Cleans the node recursively
    ///
    /// Removes the empty children nodes. The empty nodes are nodes without children and attributes.
    /// @param node CXmlNode*, the node to clean
    void clean(CXmlNode* node);

public:

    /// @brief Constructor
    ///
    /// @param fileName std::string, the registry file name w/o path
    /// @param programGroupName std::string, the name of the program group to generate a directory name for the registry files.
    /// Should be a single phrase without '\\' or '/'
    /// @param mode CRegistryMode, see CRegistryMode for details
    CRegistry(std::string fileName,std::string programGroupName,CRegistryMode mode=USER_REGISTRY);

    /// @brief Destructor
    virtual  ~CRegistry();

    /// @brief Sets the registry file name
    void fileName(std::string fname) {
        m_fileName = fname;
    }

    /// @brief Returns the registry file name
    const std::string fileName() {
        return m_fileName;
    }

    /// @brief Loads registry from the file. Usually it's the first action with registry.
    virtual void load();

    /// @brief Loads registry from the string list
    virtual void load(const CStrings& data);

    /// @brief Loads registry from XML node
    ///
    /// Any XML node that has subnodes is considered as section.
    /// Nested sections make paths with the elements separated with "/".
    /// @param data const CXmlDoc&, the XML document to load data from
    virtual void load(const CXmlDoc& data);

    /// @brief Loads registry from buffer.
    /// @param buffer const CBuffer&, source buffer
    virtual void load(const CBuffer &buffer) { clear(); CXmlDoc::load(buffer); }

    /// @brief Saves registry to the file.
    virtual void save();

    /// @brief Saves registry to the the string list
    virtual void save(CStrings& data);

    /// @brief Saves registry to XML node
    ///
    /// Nested sections with paths with the elements separated with "/" make nested XML nodes.
    /// @param data const CXmlDoc&, the XML document to load data from
    virtual void save(CXmlDoc& data) const;

    /// @brief Saves registry to buffer.
    /// @param buffer CBuffer&, a buffer to save document
    /// @param indent int, how many indent spaces at start
    virtual void save(CBuffer &buffer, int indent=0) const { CXmlDoc::save(buffer,indent); }

public:

    /// Finds out the user's home directory
    static std::string homeDirectory();
};
/// @}
}
#endif
