/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CSharedStrings.h  -  description
                             -------------------
    begin                : June 1 2006
    copyright            : (C) 2006-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CSHAREDSTRINGS_H__
#define __CSHAREDSTRINGS_H__

#include <string>
#include <map>

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

/// @brief Shared strings table
///
/// Contains a table of shared strings for the use with different objects
/// such as XML node, ListView, etc
class CSharedStrings {
    typedef std::map<std::string, int> CSIMap; ///< String to int map type
    CSIMap m_stringIdMap; ///< Map of shared strings and reference counters
public:
    /// @brief Default constructor
    CSharedStrings();

    /// @brief Destructor
    ~CSharedStrings() {
        clear();
    }

    /// @brief Obtain a shared string
    ///
    /// Looks for an existing shared string, and returns a const std::string&
    /// to it. If a shared string not found, it's created with reference count one.
    /// For an existing shared string, every call of this method encreases string
    /// reference counter by one
    /// @param str const char *, a string to share
    const std::string& shareString(const char *str);

    /// @brief Obtain a shared string
    ///
    /// Looks for an existing shared string, and returns a const char *pointer
    /// to it. If a shared string not found, it's created with reference count one.
    /// For an existing shared string, every call of this method encreases string
    /// reference counter by one
    /// @param str const std::string&, a string to share
    const std::string& shareString(const std::string& str) {
        return shareString(str.c_str());
    }

    /// @brief Releases a shared string
    ///
    /// If the shared string reference counter is greater than one,
    /// it's decreased by one. If the reference counter is one,
    /// the string is removed from SST. If the string doesn't exist,
    /// the exception is thrown.
    /// @param str const char *, a string to release
    void releaseString(const char *str) throw(std::exception);

    /// @brief Releases a shared string
    ///
    /// If the shared string reference counter is greater than one,
    /// it's decreased by one. If the reference counter is one,
    /// the string is removed from SST. If the string doesn't exist,
    /// the exception is thrown.
    /// @param str const std::string&, a string to release
    void releaseString(const std::string& str) throw(std::exception) {
        releaseString(str.c_str());
    }

    /// @brief Clears the shared string table
    ///
    /// Only the string containing the empty string ("") is left
    /// after this operation
    void clear();
}
;
/// @}
}

#endif
