/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CSocket.cpp  -  description
                             -------------------
    begin                : July 10 2002
    copyright            : (C) 2002-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CSOCKET_H__
#define __CSOCKET_H__

#ifndef _WIN32
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <sys/un.h>
    #include <unistd.h>
    #include <fcntl.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <netdb.h>

    /// A socket handle is an integer
    typedef int SOCKET;

    /// A value to indicate an invalid handle
    #define INVALID_SOCKET -1

#else
    #include <winsock2.h>
    #include <windows.h>
#endif

#include <sptk3/CException.h>
#include <sptk3/CStrings.h>
#include <sptk3/CBuffer.h>

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

/// Buffered Socket reader.
class CSocketReader : protected CBuffer {
    int       m_socket;       ///< Socket handle
    uint32_t  m_readOffset;   ///< Current offset in the read buffer

    /// @brief Performs buffered read
    ///
    /// Data is read from the opened socket into a character buffer of limited size
    /// @param dest char *, destination buffer
    /// @param sz uint32_t, size of the destination buffer
    /// @param readLine bool, true if we want to read one line (ended with CRLF) only
    /// @param from sockaddr_in*, an optional structure for source address
    /// @returns number of bytes read
    int32_t bufferedRead(char *dest,uint32_t sz,bool readLine,struct sockaddr_in* from=NULL);

public:

    /// @brief Constructor
    /// @param bufferSize int, the desirable size of the internal buffer
    CSocketReader(int bufferSize=4096);

    /// @brief Connects the reader to the socket handle
    /// @param socket int, socket handle
    void open(int socket);

    /// @brief Disconnects the reader from the socket handle
    void close() {
        m_socket = 0;
    }

    /// @brief Performs the buffered read
    /// @param dest char *, destination buffer
    /// @param sz uint32_t, size of the destination buffer
    /// @param readLine bool, true if we want to read one line (ended with CRLF) only
    /// @param from sockaddr_in*, an optional structure for source address
    /// @returns bytes read from the internal buffer
    uint32_t read(char *dest,uint32_t sz,bool readLine=false,struct sockaddr_in* from=NULL);

    /// @brief Performs the buffered read of '\n'-terminated string
    /// @param dest CBuffer&, destination buffer
    /// @returns bytes read from the internal buffer
    uint32_t readLine(CBuffer& dest);
};

/// @brief Generic socket.
///
/// Allows to establish a network connection
/// to the host by name and port address
class CSocket {
protected:
    SOCKET              m_sockfd;       ///< Socket internal (OS) handle
    int32_t             m_domain;       ///< Socket domain type
    int32_t             m_type;         ///< Socket type
    int32_t             m_protocol;     ///< Socket protocol
    std::string         m_host;         ///< Host name
    int32_t             m_port;         ///< Port number
    fd_set              m_inputs;       ///< The set of socket descriptors for reading
    fd_set              m_outputs;      ///< The set of socket descriptors for writing
    CSocketReader       m_reader;       ///< Socket buffered reader
    CBuffer             m_stringBuffer; ///< Buffer to read a line

protected:

#ifdef _WIN32
    static int32_t      m_socketCount;  ///< Total number of the objects of this class
    static bool         m_inited;       ///< Were the sockets inited?
    static void         init();         ///< WinSock initialization
    static void         cleanup();      ///< WinSock cleanup
#endif

    /// Reads a single char from the socket
    char getChar();

public:
    /// A mode to open a socket, one of
    enum CSocketOpenMode {
        SOM_CREATE,     ///< Only create (Typical UDP connectionless socket)
        SOM_CONNECT,    ///< Connect
        SOM_BIND        ///< Bind (listen)
    };

public:
    /// Constructor
    CSocket(int32_t domain=AF_INET, int32_t type=SOCK_STREAM, int32_t protocol=0);

    /// Destructor
    virtual ~CSocket();

    /// Returns socket handle
    int  handle() const {
        return (int) m_sockfd;
    }

    /// Attaches socket handle
    void attach(SOCKET socketHandle);

    /// Sets the host name
    void host(std::string hostName);

    /// Returns the host name
    std::string host() const {
        return m_host;
    }

    /// Sets the port number
    void port(int32_t portNumber);

    /// Returns the current port number
    /// @returns port number
    int32_t port() const {
        return m_port;
    }

    /// @brief Opens the socket connection by address.
    /// @param openMode CSocketOpenMode, SOM_CREATE for UDP socket, SOM_BIND for the server socket, and SOM_CONNECT for the client socket
    /// @param addr sockaddr_in*, defines socket address/port information
    void open_addr(CSocketOpenMode openMode=SOM_CREATE,sockaddr_in* addr=0L);

    /// @brief Opens the client socket connection by host and port
    virtual void open(std::string hostName="", int32_t port=0);

    /// Opens the server socket connection on port (binds/listens)
    void listen(int32_t portNumber=0);

    /// @brief In server mode, waits for the incoming connection.
    ///
    /// When incoming connection is made, exits returning the connection info
    /// @param clientSocketFD int&, connected client socket FD
    /// @param clientInfo sockaddr_in&, connected client info
    void accept(int& clientSocketFD,struct sockaddr_in& clientInfo);

    /// @brief Closes the socket connection
    virtual void close();

    /// @brief Returns the current socket state
    /// @returns true if socket is opened
    bool active() const {
        return m_sockfd != INVALID_SOCKET;
    }

    /// @brief Calls Unix fcntl() or Windows ioctlsocket()
    int32_t  control(int flag, uint32_t *check);

    /// @brief Sets socket option value
    /// Throws an error if not succeeded
    void setOption(int level,int option,int value) throw(sptk::CException);

    /// @brief Gets socket option value
    ///
    /// Throws an error if not succeeded
    void getOption(int level,int option,int& value) throw(sptk::CException);

    /// @brief Reads one line (terminated with CRLF) from the socket into existing memory buffer
    ///
    /// The output string should fit the buffer or it will be returned incomplete.
    /// @param buffer char *, the destination buffer
    /// @param size uint32_t, the destination buffer size
    /// @returns the number of bytes read from the socket
    uint32_t readLine(char *buffer,uint32_t size);

    /// @brief Reads one line (terminated with CRLF) from the socket into existing memory buffer
    ///
    /// The memory buffer is extended automatically to fit the string.
    /// @param buffer CBuffer&, the destination buffer
    /// @returns the number of bytes read from the socket
    uint32_t readLine(CBuffer& buffer);

    /// @brief Reads one line (terminated with CRLF) from the socket into string
    /// @param s std::string&, the destination string
    /// @returns the number of bytes read from the socket
    uint32_t readLine(std::string& s);

    /// @brief Reads data from the socket
    /// @param buffer char *, the memory buffer
    /// @param size uint32_t, the memory buffer size
    /// @param from sockaddr_in*, an optional structure for source address
    /// @returns the number of bytes read from the socket
    uint32_t read(char *buffer,uint32_t size,sockaddr_in* from=NULL);

    /// @brief Reads data from the socket into memory buffer
    /// @param buffer const CBuffer&, the memory buffer
    /// @param from sockaddr_in*, an optional structure for source address
    /// @returns the number of bytes read from the socket
    uint32_t read(CBuffer& buffer,sockaddr_in* from=NULL);

    /// @brief Writes data to the socket
    /// @param buffer const char *, the memory buffer
    /// @param size uint32_t, the memory buffer size
    /// @param peer const sockaddr_in*, optional peer information
    /// @returns the number of bytes written to the socket
    uint32_t write(const char *buffer,uint32_t size,const sockaddr_in* peer=0);

    /// @brief Writes data to the socket
    /// @param buffer const CBuffer&, the memory buffer
    /// @returns the number of bytes written to the socket
    uint32_t write(const CBuffer& buffer);

    /// @brief Reports true if socket is ready for reading from it
    bool readyToRead(uint32_t waitmsec);

    /// @brief Reports true if socket is ready for writing to it
    bool readyToWrite();

    /// @brief Stream std::string input
    const CSocket& operator << ( const std::string& );

    /// @brief Stream std::string output
    const CSocket& operator >> ( std::string& );
};

/// @}
}
#endif
