/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CStrings.h  -  description
                             -------------------
    begin                : Thu August 11 2005
    copyright            : (C) 2005 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CSTRINGS_H__
#define __CSTRINGS_H__

#include <sptk3/sptk.h>
#include <sptk3/string_ext.h>
#include <sptk3/CException.h>
#include <string>
#include <vector>
#include <algorithm>

namespace sptk {

/// @addtogroup utility Utility Classes
/// @{

/// A workaround for VC++ bug
typedef std::string std_string;

/// @brief String with ID
/// Extended version of std::string that supports an integer string ID
class SP_EXPORT idstring : public std_string {
public:
    int32_t       m_id;    ///< String ID
public:
    /// Default constructor
    idstring() {
        m_id = 0;
    }

    /// Copy constructor
    /// @param str const idstring&, source string
    idstring(const idstring& str) : std_string(str), m_id(str.m_id) {}

    /// Constructor
    /// @param str const string&, source string
    /// @param id in, optional string id
    idstring(const std::string& str,int32_t id=0) : std_string(str), m_id(id) {}

    /// Constructor
    /// @param str const char *, source string
    /// @param id in, optional string id
    idstring(const char *str,int32_t id=0) : std_string(str), m_id(id) {}

    /// Assignment operator
    idstring& operator = (const idstring& si)  {
        assign(si);
        m_id = si.m_id;
        return *this;
    }

    /// Assignment operator
    /// @param str const char *, source string
    idstring& operator = (const char * str) {
        assign(str);
        m_id = 0;
        return *this;
    }

    /// Returns string ID
    int32_t ident() const {
        return m_id;
    }

    /// Sets string ID
    void ident(int32_t id) {
        m_id = id;
    }
};

/// @brief List of strings with ids
///
/// General string list. Based on vector<idstring>. Stores strings with (optional) integer Ids.
/// Includes several extra methods to construct it from string or load/save from/to file.
class CStrings : public std::vector<idstring> {

    int32_t m_userData;    ///< User-specified data

    /// Parses a string to elements separated by a delimiter string
    /// @param src const char *, a source string
    /// @param separator const char *, a separator string
    void parseString(const char *src,const char *delimitter);
public:
    /// @brief Default constructor
    CStrings() {
        m_userData = 0;
    }

    /// @brief Copy constructor
    CStrings(const CStrings& src) : std::vector<idstring>() {
        resize(src.size());
        std::copy(src.begin(),src.end(),begin());
    }

    /// @brief Constructor from a string with elements separated by a delimiter string
    /// @param src const std::string&, a source string
    /// @param separator const char *, a separator string
    CStrings(const std::string& src,const char *separator) {
        parseString(src.c_str(),separator);
    }

    /// @brief Constructor from a string with elements separated by a delimiter string
    /// @param src const char *, a source string
    /// @param separator const char *, a separator string
    CStrings(const char *src,const char *separator) {
        clear();
        parseString(src,separator);
    }

    /// @brief Assigns strings from a string with elements separated by a delimiter string
    /// @param src const std::string&, a source string
    /// @param separator const char *, a separator string
    void fromString(const std::string& src,const char *separator) {
        clear();
        parseString(src.c_str(),separator);
    }

    /// @brief Makes string from own strings separated by a delimiter string
    /// @param separator const char *, a separator string
    std::string asString(const char *separator) const;

    /// @brief Returns an index of the string in strings, or -1 if not found
    /// @param s std::string, a string to find
    /// @returns a string index, or -1
    int indexOf(std::string s) const;

    /// @brief Saves strings to file. String ids are discarded.
    /// @param fileName std::string, the name of the file
    void saveToFile(std::string fileName) const throw(CException);

    /// @brief Loads strings from file. String ids are not loaded.
    /// @param fileName std::string, the name of the file
    void loadFromFile(std::string fileName) throw(CException);

    /// @brief Returns user data as integer
    int32_t argument() const   {
        return (int)m_userData;
    }

    /// @brief Sets user data as integer
    /// @param d int, new value for user data
    void argument(int32_t d)   {
        m_userData = d;
    }

    /// @brief Removes a string from vector
    /// @param i uint32_t, string index in the vector
    void remove(uint32_t i) {
        erase(begin()+i);
    }
}
;
/// @}
}
#endif
