/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CExcelBook.h  -  description
                             -------------------
    begin                : Mon Jan 19 2004
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CEXCELBOOK_H__
#define __CEXCELBOOK_H__

#include <sptk3/excel/CExcelPalette.h>
#include <sptk3/excel/CExcelCellFormat.h>
#include <sptk3/excel/CExcelSheet.h>
#include <sptk3/excel/CExcelStyle.h>
#include <sptk3/excel/CExcelHeader.h>
#include <sptk3/excel/CExcelBAT.h>
#include <sptk3/excel/CExcelProperty.h>
#include <sptk3/excel/CExcelSST.h>
#include <sptk3/CStrings.h>
#include <stdio.h>

namespace sptk {

/// @addtogroup excel Excel Classes
/// @{

/// Excel book reading/writing progression callback
typedef void (*CExcelBookCallback)(const char *operationName,int percentage);

/// @brief The Excel book
class SP_EXPORT CExcelBook {
    friend class CExcelSheet;
protected:
    std::string           m_name;              ///< Abstract workbook name
    std::string           m_lastError;         ///< Last processing error
    int32_t               m_codePage;          ///< Document codepage
    int32_t               m_codePageConvert;   ///< Codepage for converting unicode to
    int16_t               m_BIFFversion;       ///< Excel BIFF format version
    CExcelRecordList      m_records;           ///< The list of records from Excel file
    uint32_t*               m_sstRecordOffsets;  ///< Excel record offsets
    uint32_t                m_sstRecordCount;    ///< Excel record count
    uint32_t                m_bofPosition;       ///< The position of the first BOF in the file
    CExcelBookCallback    m_callback;          ///< Load/save progression callback
    CExcelStyles          m_styles;            ///< The list of global styles

    /// @brief Computes the number of blocks (rounded up) using file positions
    /// @param start uint32_t, the start position in the file
    /// @param end uint32_t, the end position in the file
    /// @param blockSize uint32_t, the size of the block in the file
    uint32_t blockCount(uint32_t start,uint32_t end,uint32_t blockSize=BIG_BLOCK_SIZE);

    /// @brief Processes one Excel record
    void process_item(const CExcelRecord& record);

    /// @brief Parses Excel File
    int  scanExcelFile(FILE *f);

    /// @brief Parses Excel File for sheet information
    uint32_t  scanSheetsInfo(FILE *f);

    /// Parses Workbook Globals
    uint32_t  parseWorkBookGlobals(uint32_t& dataStart);

    /// @brief Decodes Excel shared string record
    uint32_t decodeSSTrecord(const CExcelRecord& record);

    /// @brief Decodes Excel extended shared string record
    uint32_t decodeEXTSSTrecord(const CExcelRecord& record);

    /// @brief Saves Excel records for writing into the CExcelRecords
    /// @param records CExcelRecords&, the output set of records
    void saveRecords(CExcelRecordsBuffer& records);
protected:
    CExcelHeader          m_excelHeader;      ///< Excel file header
    CExcelBAT             m_excelBAT;         ///< Excel file Block Allocation Table (BAT)
    CExcelBAT             m_excelSBAT;        ///< Excel file Small Block Allocation Table (SBAT)
    CExcelProperties      m_excelProperties;  ///< Excel file Properties
    CBuffer              *m_workBookData;     ///< Workbook records buffer

    /// @brief Calls a callback function (if set)
    void do_callback(const char *operation,int percentage);

	/// @brief Converts unicode string to buffer
	/// @param dest CBuffer&, the destination buffer
	/// @param ptr const char*, the input string
	/// @param len uint32_t, the string len in characters
	/// @returns number of bytes in the input string
	uint32_t unicodeToString(CBuffer& dest, const char *ptr,uint32_t len);

public:
    CExcelSheets          sheets;             ///< The list of spreadsheets
    CExcelSST             sharedStringTable;  ///< The list of shared strings
    CExcelCellFormats     xfRecords;          ///< The list of formats
    CExcelPalette         palette;            ///< The list of the available colours

    /// @brief Constructor
    /// @param name CString, an abstract name to associate with the workbook
    CExcelBook(std::string name="");

    /// @brief Destructor
    ~CExcelBook();

    /// @brief Clears inside structure completely
    void clear();

    /// @brief Clears inside structure completely
    ///
    /// Also creates the minimal set of fonts, formats, etc.
    void init();

    /// @brief Returns an abstract workbook name
    const std::string& name() const {
        return m_name;
    }

    /// @brief Sets an abstract workbook name
    void name(const char *nm) {
        m_name = nm;
    }

    /// @brief Reads the Excel file
    /// @param fileName CString, the name of the input file
    int readExcelFile(std::string fileName);

    /// @brief Reads the CSV file
    /// The sheet is created automatically for every imported CSV file
    /// @param fileName CString, the name of the input file
    int readCSVFile(std::string fileName);

    /// @brief Saves the Excel file
    /// @param fileName CString, the name of the output file
    int writeExcelFile(std::string fileName);

    /// @brief Saves the Excel file
    /// @param file FILE *, the external output file
    int writeExcelFile(FILE *file);

    /// @brief Saves the CSV file
    /// @param fileName CString, the name of the output file
    /// @param sheetNumber uint32_t, the number of sheet to export (the default is 0, the very first sheet)
    int writeCSVFile(std::string fileName,uint32_t sheetNumber=0);

    /// @brief Saves the CSV file
    /// @param file FILE *, the external output file
    /// @param sheetNumber uint32_t, the number of sheet to export (the default is 0, the very first sheet)
    int writeCSVFile(FILE *file,uint32_t sheetNumber=0);

    /// @brief Reports the last error
    const char * lastError() {
        return m_lastError.c_str();
    }

    /// @brief Returns load/save progression callback
    CExcelBookCallback callback() const {
        return m_callback;
    }

    /// @brief Defines load/save progression callback
    /// @param cb CExcelBookCallback, the new callback function
    void callback(CExcelBookCallback cb) {
        m_callback = cb;
    }

    /// @brief Returns a numeric format for the index
    /// @param index uint32_t, logical index for the numeric format
    const CExcelNumericFormat& numericFormat(uint32_t index) const {
        return xfRecords.numericFormat(index);
    }

    /// @brief Creates a new sheet in the book
    CExcelSheet *newSheet(const char *sheetName);

    /// @brief Adds a new font to the font list, or finds similar existing font
    /// @param fontName const char *, font name
    /// @param fontHeight uint32_t, font height in points
    /// @param fontAttr uint32_t, font attr (EFM_ATTR_ITALIC,EFM_ATTR_BOLD,EFM_ATTR_UNDERLINE,EFM_ATTR_STRIKEOUT)
    /// @param charSet uint32_t, character set as defined in Windows API LOGFONT structure
    /// @returns logical index of the new or existing font
    uint32_t defineFont(const char *fontName,uint32_t fontHeight,uint16_t fontAttr=0,uint32_t charSet=0);

    /// @brief Returns the font by the logical font index.
    /// Font with index 4 is a dummy font and shouldn't be used.
    const CExcelFont& font(uint32_t fontIndex) const;

    /// @brief Defines a new cell format from the scratch
    /// @param fontIndex uint32_t, the logical index of the existing font
    /// @param numericFormatIndex uint32_t, the logical index of the existing numeric format
    /// @param align uint16_t, the combination of align flags from enum CExcelCellAlignFlags
    /// @param wrap bool, the cell text wrap flag
    /// @param border CExcelBorderLine, the type of the cell border line
    /// @param pattern uint32_t, one of the Excel standard pattern types 0..0x12
    /// @param patternForeColorIndex uint32_t, the pattern fore color index in the palette
    /// @param patternBackColorIndex uint32_t, the pattern back color index in the palette
    /// @returns the pointer on the new cell format
    CExcelCellFormat *defineFormat(
        uint32_t fontIndex=0,
        uint32_t numericFormatIndex=0,
        uint16_t align=0,
        bool     wrap=false,
        CExcelBorderLine border=CBL_NONE,
        uint32_t pattern=0,
        uint32_t patternForeColorIndex=EPC_BLACK,
        uint32_t patternBackColorIndex=EPC_WHITE
    );

	/// @brief Returns the code page used for conversion from unicode
	///
	/// The default is the current system code page
	int codePageConvert() const { return m_codePageConvert; }

	/// @brief Sets the code page used for conversion from unicode
	///
	/// The default is the current system code page
	/// @param codePage int, the code page
	void codePageConvert(int codePage) { m_codePageConvert = codePage; }
};
/// @}
}
#endif
