/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CExcelCellFormat.h  -  description
                             -------------------
    begin                : Mon Jan 19 2004
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CEXCELFORMATRECORD_H__
#define __CEXCELFORMATRECORD_H__

#include <sptk3/CVariant.h>
#include <sptk3/CException.h>
#include <sptk3/excel/CExcelFont.h>
#include <sptk3/excel/CExcelNumericFormat.h>
#include <sptk3/excel/CExcelPalette.h>

#include <vector>

namespace sptk {

/// @addtogroup excel Excel Classes
/// @{

#define MAX_BIFF8FORMAT 0x31    ///< Max BIFF8 format

/// Possible align format flags for cell formatting. These flags
/// can be combined to get the desired alignment using bit operations.
enum CExcelCellAlignFlags {
    CAF_GENERAL=0,    ///< Default alignment
    CAF_LEFT,         ///< Left align text horizontally
    CAF_CENTER,       ///< Center text horizontally
    CAF_RIGHT,        ///< Right align text horizontally
    CAF_FILL,         ///< Fill text horizontally
    CAF_JUSTIFY,      ///< Justify text horizontally
    CAF_WRAP=8,       ///< Wrap text
    CAF_VCENTER=9,    ///< Center text vertically
    CAF_VBOTTOM=10,   ///< Align text to the bottom
    CAF_VJUSTIFY=11   ///< Justify text vertically
};

/// Possible border types for the cell
enum CExcelBorderLine {
    CBL_NONE,       ///< No border
    CBL_THIN,       ///< Thin line
    CBL_MEDIUM,     ///< Medium line
    CBL_DASHED,     ///< Dashed line
    CBL_DOTTED,     ///< Dotted line
    CBL_THICK,      ///< Thick line
    CBL_DOUBLE,     ///< Double line
    CBL_HAIR,       ///< Very thin line
    CBL_MDASHED,    ///< Medium dashed line
    CBL_DASHDOT,    ///< Dash-dot line
    CBL_MDASHDOT,   ///< Medium dash-dot line
    CBL_DASHDOTDOT, ///< Dash-dot-dot line
    CBL_MDASHDOTDOT,///< Medium dash-dot-dot line
    CBL_SDASHDOT    ///< Slanted dash-dot line
};

/// The format descriptor
typedef struct ifmt {
    CVariantType   kind;           ///< Format data type
    unsigned char  id;             ///< Format ID
    const char     *format;        ///< Format string
}
IFMT; ///< The format descriptor type

/// Class CExcelCellFormat describes the cell data presentation.
class SP_EXPORT CExcelCellFormat {
    friend class CExcelCellFormats;
protected:
    char                 m_data[20];           ///< Internal data buffer
    uint32_t             m_index;              ///< Index of this record in XF records table
    CExcelNumericFormat *m_dataFormat;         ///< Cell data format information

protected:
    /// Protected constructor, builds a new format record from the scratch
    /// @param fontIndex uint32_t, the logical index of the existing font
    /// @param numericFormatIndex uint32_t, the logical index of the existing numeric format
    /// @param align uint32_t, the combination of align flags from enum CExcelCellAlignFlags
    /// @param wrap bool, the cell text wrap flag
    /// @param border CExcelBorderLine, the type of the cell border line
    /// @param pattern uint32_t, one of the Excel standard pattern types 0..0x12
    /// @param patternForeColorIndex uint32_t, the pattern fore color index in the palette
    /// @param patternBackColorIndex uint32_t, the pattern back color index in the palette
    CExcelCellFormat(
        uint32_t   fontIndex=0,
        uint32_t   numericFormatIndex=0,
        uint16_t align=0,
        bool     wrap=false,
        CExcelBorderLine border=CBL_NONE,
        uint32_t pattern=0,
        uint32_t patternForeColorIndex=EPC_BLACK,
        uint32_t patternBackColorIndex=EPC_WHITE
    );

    /// Protected constructor, builds a new format record from the existing XF record data
    /// @param XFData const void *, the XF record data to be copied inside
    CExcelCellFormat(const void *XFData);

    /// Protected copy constructor
    /// @param xfRecord const CExcelCellFormat&, the XF record
    CExcelCellFormat(const CExcelCellFormat& xfRecord);

public:

    /// Returns the index of this record in XF records table
    uint32_t index() const {
        return m_index;
    }

    /// Returns a pointer to data buffer that contains XF record data
    const char *data() const {
        return m_data;
    }

    /// Returns font index in the fonts table
    uint32_t fontIndex() const {
        return CExcelRecord::getshort(m_data,0);
    }

    /// Sets font index from the fonts table
    /// @param fi uint32_t, font index from the fonts table
    void fontIndex(uint32_t fi) {
        CExcelRecord::setshort(m_data,0,(short)fi);
    }

    /// Returns format index in the formats table
    uint32_t formatIndex() const {
        return CExcelRecord::getshort(m_data,2);
    }

    /// Sets format index in the formats table
    void formatIndex(uint32_t fi) {
        CExcelRecord::setshort(m_data,2,(uint16_t)fi);
    }

    /// Retuns the align flags for the cell
    uint16_t align() const {
        return CExcelRecord::getshort(m_data,6)&0x77;
    }

    /// Sets the align flags for the cell
    void align(uint16_t flags);

    /// Retuns the wrap flag for the cell
    uint32_t wrap() const {
        return CExcelRecord::getshort(m_data,6)&0x8;
    }

    /// Sets the wrap flag for the cell
    void wrap(bool flag);

    /// Returns border line style for the left edge
    CExcelBorderLine border() const  {
        return CExcelBorderLine(CExcelRecord::getshort(m_data,10) & 0xF);
    }

    /// Sets border line style
    /// @param ls CExcelBorderLines, border line style
    void border(CExcelBorderLine ls);

    /// Returns the index of the pattern background color in the palette
    uint32_t patternBackColor() const;

    /// Sets the index of the pattern background color in the palette
    /// @param patternBackColorIndex uint32_t, index of color in the palette
    void patternBackColor(uint32_t patternBackColorIndex);

    /// Returns the index of the pattern foreground color in the palette
    uint32_t patternForeColor() const;

    /// Sets the index of the pattern foreground color in the palette
    /// @param patternForeColorIndex uint32_t, index of color in the palette
    void patternForeColor(uint32_t patternForeColorIndex);

    /// Defines fill pattern index for the background
    /// @param patternIndex uint32_t, the index of the standard pattern
    void fillPattern(uint32_t patternIndex);

    /// Returns numeric format info
    const CExcelNumericFormat& dataFormat() const {
        return *m_dataFormat;
    }
};

/// The list of the available cell formats.
class CExcelCellFormats {
    friend class CExcelBook;
protected:
    CExcelNumericFormats               m_numericFormats;   ///< The list of the available numeric formats
    std::vector<CExcelCellFormat *>    m_formatList;       ///< The list of the XF formats
    CExcelFonts                        m_fonts;            ///< The list of the available fonts

public:
    /// Default constructor
    CExcelCellFormats()  {
        init();
    }

    /// Destructor
    ~CExcelCellFormats() {
        clear();
    }

    /// Adds a font to the font list, or finds similar existing font
    /// @param fontName const char *, font name
    /// @param fontHeight uint32_t, font height in points
    /// @param fontAttr uint32_t, font attr (EFM_ATTR_ITALIC,EFM_ATTR_STRIKEOUT)
    /// @param charSet uint32_t, character set as defined in Windows API LOGFONT structure
    /// @returns logical index of the new or existing font
    uint32_t addFont(const char *fontName,uint32_t fontHeight,uint16_t fontAttr=0,uint32_t charSet=0);

    /// Adds a font to the font list, even if similar font exists. In the last case the metrics field
    /// is appended with the logical index of the font. That allows to read the Excel files with several
    /// identical FONT records defined
    /// @param fontDataSize uint32_t, the size of the FONT record data
    /// @param fontData const char *, FONT record data in internal Excel format
    /// @returns logical index of the new font
    uint32_t addFont(uint16_t fontDataSize,const char *fontData);

    /// Adds a new numeric format. Any attempts to re-define the built-in Excel Formats are ignored.
    /// @param numericFormatIndex uint32_t, numeric format index
    /// @param numericFormat CString, numeric format
    void addNumericFormat(uint32_t numericFormatIndex,std::string numericFormat);

    /// Inserts new format.
    /// The format information is kept as a pointer.
    /// The associated memory will be relesed upon clear() or destructor.
    /// @param xfr CExcelCellFormat *, the format record
    CExcelCellFormat *insert(CExcelCellFormat *xfr);

    /// Inserts new format from the existing Excel XF record
    /// @param XFData const void *, the format record in internal XF format
    CExcelCellFormat *insert(const void *XFData);

    /// Inserts new format, builds a new format record from the scratch
    /// @param fontIndex uint32_t, the logical index of the existing font
    /// @param numericFormatIndex uint32_t, the logical index of the existing numeric format
    /// @param align uint32_t, the combination of align flags from enum CExcelCellAlignFlags
    /// @param wrap bool, the cell text wrap flag
    /// @param border CExcelBorderLine, the type of the cell border line
    /// @param pattern uint32_t, one of the Excel standard pattern types 0..0x12
    /// @param patternForeColorIndex uint32_t, the pattern fore color index in the palette
    /// @param patternBackColorIndex uint32_t, the pattern back color index in the palette
    CExcelCellFormat *insert(
        uint32_t   fontIndex=0,
        uint32_t   numericFormatIndex=0,
        uint16_t align=0,
        bool     wrap=false,
        CExcelBorderLine border=CBL_NONE,
        uint32_t pattern=0,
        uint32_t patternForeColorIndex=EPC_BLACK,
        uint32_t patternBackColorIndex=EPC_WHITE
    );

    /// Clears the list of numeric formats, fonts, cell formats
    void clear();

    /// Clears the list of numeric formats, fonts, cell formats,
    /// then creates the minimal set of formats, fonts, and cell formats
    void init();

    /// Format access by format index
    /// @param i uint32_t, the format index
    const CExcelCellFormat& operator [] (uint32_t i) const {
        return *m_formatList[i];
    }

    /// Returns the number of available formats
    uint32_t size() const {
        return (uint32_t) m_formatList.size();
    }

    /// Encodes FORMAT and XF into internal Excel format
    void saveRecords(CExcelRecordsBuffer& records);

    /// Returns a numeric format for the index
    /// @param index uint32_t, logical index for the numeric format
    const CExcelNumericFormat& numericFormat(uint32_t index) const {
        CExcelNumericFormats::const_iterator itor = m_numericFormats.find(index);
        if (itor != m_numericFormats.end())
            return *itor->second;
        throw CException("Invalid (unknown) numeric format",__FILE__,__LINE__);
    }

    /// Returns a font for the index
    /// @param index uint32_t, logical index for the font
    const CExcelFont& font(uint32_t index) const {
        return m_fonts[index];
    }
};
/// @}
}
#endif
