/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CExcelFont.h  -  description
                             -------------------
    begin                : Mon Jan 19 2004
    copyright            : (C) 2004 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CEXCELFONT_H__
#define __CEXCELFONT_H__

#include <sptk3/excel/CExcelRecord.h>
#include <sptk3/excel/CExcelRecordsBuffer.h>

#include <sptk3/CStrings.h>
#include <map>

namespace sptk {

#define EFM_ATTR_BOLD      1
#define EFM_ATTR_ITALIC    2
#define EFM_ATTR_UNDERLINE 4
#define EFM_ATTR_STRIKEOUT 8

/// @addtogroup excel Excel Classes
/// @{

/// Excel FONT record
class CExcelFont : public CExcelRecord {
    friend class CExcelFonts;
protected:
    std::string   m_fontMetrics;  ///< A string that uniquely identifies the font - font metrics
    uint32_t        m_fontIndex;    ///< Logical index of the font
protected:
    /// Protected constructor
    /// @param fontIndex uint32_t, logical index of the font
    /// @param fontName const char *, font name
    /// @param fontHeight uint32_t, font height in points
    /// @param fontAttr uint32_t, font attr (EFM_ATTR_ITALIC,EFM_ATTR_BOLD,EFM_ATTR_UNDERLINE,EFM_ATTR_STRIKEOUT)
    /// @param charSet uint32_t, character set as defined in Windows API LOGFONT structure
    CExcelFont(uint32_t fontIndex,const char *fontName,uint32_t fontHeight,uint16_t fontAttr=0,uint32_t charSet=0);

    /// Protected constructor
    /// @param fontIndex uint32_t, logical index of the font
    /// @param fontDataSize uint16_t, the size of the FONT record data
    /// @param fontData const char *, FONT record data in internal Excel format
    CExcelFont(uint32_t fontIndex,uint16_t fontDataSize,const char *fontData);

public:
    std::string  name() const;     ///< Returns font name
    uint32_t height() const;       ///< Returns font height
    uint16_t attr() const;         ///< Returns font attributes

    /// Sets font name
    /// @param fname std::string, font name i.e. "Arial"
    void name(std::string fname);

    /// Sets font height
    /// @param fheight uint32_t, font height in points
    void height(uint32_t fheight);

    /// Sets font attributes
    /// @param fattr uint16_t, font attr (EFM_ATTR_ITALIC,EFM_ATTR_STRIKEOUT)
    void attr(uint16_t fattr);

    /// Returns a string uniquely identifying font - font metrics
    std::string fontMetrics() const {
        return m_fontMetrics;
    }

    /// Returns logical index of the font
    uint32_t index() const {
        return m_fontIndex;
    }

    /// Computes the unique string defining this font
    /// @param fontName const char *, font name
    /// @param fontHeight uint32_t, font height
    /// @param fontAttr uint32_t, font attributes
    /// @param charSet uint32_t, the character set
    static std::string computeFontMetrics(const char *fontName,uint32_t fontHeight,uint32_t fontAttr,uint32_t charSet);
};

/// @brief The list of Excel file fonts
class CExcelFonts {
    typedef std::map<std::string,CExcelFont *> CFontNamesMap;
    typedef std::vector<CExcelFont *> CFontList;
    CFontList                       m_fontList;	///< Fonst list
    CFontNamesMap                   m_fontMap;	///< Map of fonts singnatures to fonts
public:
    /// Default constructor
    CExcelFonts() {}

    /// Creates the minimal set of required fonts
    void init();

    /// Encodes fonts into internal Excel format
    void saveRecords(CExcelRecordsBuffer& records);

    /// Adds a font to the font list, or finds similar existing font
    /// @param fontName const char *, font name
    /// @param fontHeight uint32_t, font height in points
    /// @param fontAttr uint16_t, font attr (EFM_ATTR_ITALIC,EFM_ATTR_BOLD,EFM_ATTR_UNDERLINE,EFM_ATTR_STRIKEOUT)
    /// @param charSet uint32_t, character set as defined in Windows API LOGFONT structure
    /// @returns logical index of the font
    uint32_t add(const char *fontName,uint32_t fontHeight,uint16_t fontAttr=0,uint32_t charSet=0);

    /// Adds a font to the font list, even if similar font exists. In the last case the metrics field
    /// is appended with the logical index of the font. That allows to read the Excel files with several
    /// identical FONT records defined
    /// @param fontDataSize uint32_t, the size of the FONT record data
    /// @param fontData const char *, FONT record data in internal Excel format
    /// @returns logical index of the new font
    uint32_t add(uint16_t fontDataSize,const char *fontData);

    /// Clears the list of the existing fonts
    void clear();

    /// Returns the number of fonts in the list
    uint32_t size() const {
        return (uint32_t) m_fontList.size();
    }

    /// Returns the font by the logical index.
    /// Index 4 is reserved and shouldn't be used
    CExcelFont& operator [] (uint32_t i) {
        return *m_fontList[i];
    }

    /// Returns the font by the logical index (const version).
    /// Index 4 is reserved and shouldn't be used
    const CExcelFont& operator [] (uint32_t i) const {
        return *m_fontList[i];
    }
};
/// @}
}
#endif
