/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CExcelProperty.h  -  description
                             -------------------
    begin                : Mon Jan 19 2004
    copyright            : (C) 2004 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CEXCELPROPERTY_H__
#define __CEXCELPROPERTY_H__

#include <stdio.h>
#include <string.h>

#include <sptk3/excel/CExcelBAT.h>

#include <vector>

namespace sptk {

/// @addtogroup excel Excel Classes
/// @{

#define PROPERTY_SIZE               0x80 ///< Property size, bytes

#define _NAME_SIZE_OFFSET           0x40 ///< Name size offset, bytes
#define _NO_INDEX                   -1   ///< No index mark

// useful offsets
#define _PROPERTY_TYPE_OFFSET       0x42 ///< Property type offset, bytes
#define _NODE_COLOR_OFFSET          0x43 ///< Node color offset, bytes
#define _PREVIOUS_PROPERTY_OFFSET   0x44 ///< Previous property offset, bytes
#define _NEXT_PROPERTY_OFFSET       0x48 ///< Next property offset, bytes
#define _CHILD_PROPERTY_OFFSET      0x4C ///< Child property offset, bytes
#define _STORAGE_CLSID_OFFSET       0x50 ///< Storage CLSID offset, bytes
#define _USER_FLAGS_OFFSET          0x60 ///< User flags offset, bytes
#define _SECONDS_1_OFFSET           0x64 ///< Seconds 1 offset, bytes
#define _DAYS_1_OFFSET              0x68 ///< Days 1 offset, bytes
#define _SECONDS_2_OFFSET           0x6C ///< Seconds 2 offset, bytes
#define _DAYS_2_OFFSET              0x70 ///< Days 2 offset, bytes
#define _START_BLOCK_OFFSET         0x74 ///< Start block offset, bytes
#define _SIZE_OFFSET                0x78 ///< Size offset offset, bytes

// node colors
#define _NODE_BLACK                 1    ///< Node color is BLACK
#define _NODE_RED                   0    ///< Node color is RED

/// @brief Represents an internal file system property in Excel File
class CExcelProperty {
    char    m_data[PROPERTY_SIZE];         ///< Property data buffer
    CBuffer m_streamData;                  ///< Property stream buffer
public:
    /// @brief Constructor
    /// @param pname CString, property name
    /// @param ptype uint32_t, property type: 1 (directory), 2 (file), or 5 (root entry)
    /// @param nodeColor uint32_t, node color: 0 (red) or 1 (black)
    /// @param previousProperty uint32_t, previous property index
    /// @param nextProperty uint32_t, next property index
    /// @param firstChildPropety uint32_t, first child property index
    /// @param startBlock uint32_t, first block of the file described by this property
    /// @param fileLength uint32_t, the actual length of the file described by this property
    CExcelProperty(
        std::string pname,
        uint32_t ptype=0x5,
        uint32_t nodeColor=0x0,
        uint32_t previousProperty=0xFFFFFFFF,
        uint32_t nextProperty=0xFFFFFFFF,
        uint32_t firstChildPropety=0xFFFFFFFF,
        uint32_t startBlock=0xFFFFFFFE,
        uint32_t fileLength=0
    );

    /// @brief Constructor
    ///
    /// Creates property from internal Excel file' binary presentation
    /// @param data const char *, the bynary presentation of the property in internal Excel format
    CExcelProperty(const char *data);

    /// @brief Returns the property name
    std::string name() const;

    /// @brief Returns property type (1-directory, 2-file, 5-root entry)
    uint32_t propertyType() const;

    /// @brief Returns binary presentation of the property
    const char *data() const {
        return m_data;
    }

    /// @brief Returns property stream buffer
    CBuffer& streamData() {
        return m_streamData;
    }

    /// @brief Finds the stream data connected to the property and loads them into m_streamData
    /// @param batTable CExcelBAT *, the existing BAT table
    /// @param f FILE *, input file
    void extract(CExcelBAT *batTable,FILE *f);

    /// @brief Returns the first block number for this property' stream in the BAT
    uint32_t firstBATBlock() const;

    /// @brief Sets the first block number for this property' stream in the BAT
    /// @param sbo uint32_t, the first block number
    void firstBATBlock(uint32_t sbo);

    /// @brief Returns actual property' stream length
    uint32_t fileLength() const;

    /// @brief Sets actual property' stream length
    void fileLength(uint32_t fl);

    /// @brief Returns block count required to store the property' stream
    uint32_t blockCount() const;
};

/// @brief The collection of properties
class CExcelProperties : public std::vector<CExcelProperty *> {
public:
    /// @brief Default constructor
    CExcelProperties() {}

    /// @brief Destructor
    ~CExcelProperties() {
        clear();
    }

    /// @brief Cleaner
    void clear();

    /// @brief Saves properties to file
    void save(FILE *f);

    /// @brief Returns number of blocks required to store the properties
    uint32_t blocks() const;

    /// @brief Extracts properties from the file and header
    /// @param fileHeader CExcelHeader *, the Excel file header
    /// @param batTable CExcelBAT *, the Excel BAT table
    /// @param f FILE *, the input file
    void extract(CExcelHeader *fileHeader,CExcelBAT *batTable,FILE *f);

    /// @brief Finds a property by name
    /// @param name const char *, property name
    CExcelProperty *find(const char *name);
};
/// @}
}
#endif
