/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CExcelSheet.h  -  description
                             -------------------
    begin                : Mon Jan 19 2004
    copyright            : (C) 2004 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CEXCELSHEET_H__
#define __CEXCELSHEET_H__

#include <stdio.h>

#include <sptk3/excel/CExcelRows.h>
#include <sptk3/excel/CExcelRecordsBuffer.h>
#include <sptk3/excel/CExcelCellFormat.h>
#include <sptk3/excel/CExcelColumn.h>

#include <vector>

namespace sptk {

/// @addtogroup excel Excel Classes
/// @{

class CExcelBook;

/// Class CExcelSheet represents an Excel spreadsheet
class SP_EXPORT CExcelSheet {
    friend class CExcelBook;
private:
    std::string    m_sheetName;             ///< Spreadsheet name
    char           m_format_buffer[64];     ///< Spreadsheet formats
    const char    *m_lastError;             ///< Last processing error
    CExcelBook    *m_parentBook;            ///< Spreadsheet's parent book
    CExcelRow     *m_currentRow;            ///< Current row in the dataset.
    short          m_BIFFversion;           ///< Excel format version
    CExcelColumns  m_columns;               ///< Excel columns info
    uint32_t         m_lastComputedPosition;  ///< Last computed position

    /// Processes an Excel file record
    void        process_item (int rectype, int reclen, char *rec);

    /// Formats the floating point data with the format ixf
    char *      format_double(double d,int ixf);
protected:
    uint32_t               m_sheetPosition;       ///< Spreadsheet position
    CExcelRecordsBuffer  m_recordsBuffer;       ///< Internal buffer for ExcelRecords

    /// Encodes sheet's data into internal records buffer
    /// The resulting recordset is used later to save sheet into Excel file.
    /// @param sheetStreamPosition uint32_t, the position of the sheet BOF in the stream
    void encode(uint32_t sheetStreamPosition);

    /// Returns internal recordsBuffer
    const CExcelRecordsBuffer& recordsBuffer() const {
        return m_recordsBuffer;
    }

    /// Reads the sheet from CSV file
    /// @param fileName CString, the name of the input file
    int readCSVFile(std::string fileName);

    /// Saves the sheet into CSV file
    /// @param fileName CString, the name of the output file
    int writeCSVFile(std::string fileName);

    /// Saves the sheet into CSV file
    /// @param file FILE *, the external output file
    int writeCSVFile(FILE *file);

    /// Loads spreadsheet from the Excel file
    int parseWorkSheet(CBuffer& excelData,uint32_t& pos);

public:
    /// Spreadsheet rows
    CExcelRows rows;

protected:
    /// Protected constructor, only CExcelBook can create new CExcelSheet
    /// @param name const char *, spreadsheet name
    /// @param filePosition uint32_t, spreadsheet position in Excel file
    /// @param parentBook CExcelBook *, parent Excel book
    CExcelSheet(const char *name,uint32_t filePosition,CExcelBook *parentBook);

public:
    /// Returns spreadsheet name
    const std::string& name() const {
        return m_sheetName;
    }

    /// Sets spreadsheet name
    void name(const char *nm) {
        m_sheetName = nm;
    }

    /// Spreadsheet's parent book
    const CExcelBook& book() const {
        return *m_parentBook;
    }

    /// Clears the spreadsheet data
    void clearData();

    /// Returns the width of the column
    /// @param col uint32_t, column number
    /// @returns the width of the column - in characters of default font
    uint32_t colWidth(uint32_t col) const {
        return m_columns[col]->width();
    }

    /// Sets the width of the column
    /// @param col uint32_t, column number
    /// @param w uint16_t, the width of the column - in characters of default font
    void colWidth(uint32_t col,uint16_t w) {
        m_columns[col]->width(w);
    }

    /// Returns the number of the rows in spreadsheet
    uint32_t rowCount() const {
        return rows.size();
    }

    /// Returns the number of the columns in spreadsheet
    uint32_t colCount() const {
        return rows[0].size();
    }

    /// Sets the data in the cell
    /// @param row uint32_t, cell row number
    /// @param col uint32_t, cell column number
    /// @param value CVariant, cell value
    /// @param format CExcelFormat *, pointer to the format record
    int setValue(uint32_t row,uint32_t col,CVariant value,const CExcelCellFormat *format);

    /// Sets the RK (integer,float,date,datetime) data in the cell
    /// @param row uint32_t, cell row number
    /// @param col uint32_t, cell column number
    /// @param value double, cell value
    /// @param rkType CVariantType, number type (integer,float)
    /// @param format CExcelFormat *, pointer to the format record
    int setRKValue(uint32_t row,uint32_t col,double value,CVariantType rkType,const CExcelCellFormat *format);

    /// Returns spreadsheet position in Excel file
    uint32_t filePosition() const {
        return m_sheetPosition;
    }

    /// Changes the size of the spreadsheet
    void resize(uint32_t height,uint32_t width);

    /// Sets format to a group of cells
    /// @param format const CExcelCellFormat *, the cell format
    /// @param row1 uint32_t, the first row in the cell group
    /// @param col1 uint32_t, the first column in the cell group
    /// @param row2 uint32_t, the last row in the cell group
    /// @param col2 uint32_t, the last column in the cell group
    void format(const CExcelCellFormat *format,uint32_t row1,uint32_t col1,uint32_t row2,uint32_t col2);

    /// Updates vertical positions for the sheet rows
    /// Uses m_lastComputedPosition to start computations from
    void updatePositions();
};

/// The list of Excel spreadsheets
class SP_EXPORT CExcelSheets : public std::vector<CExcelSheet *> {
    CExcelBook  *m_parentBook; ///< Parent Excel book
public:

    /// Default constructor
    CExcelSheets() {
        m_parentBook = 0L;
    }

    /// Destructor
    ~CExcelSheets() {
        clear();
    }

    /// Cleaner
    void clear();

    /// Attach the list to the parent Excel book
    void attach(CExcelBook *book) {
        m_parentBook = book;
    }
};
/// @}
}
#endif
