/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          cbuttongroup.h  -  description
                             -------------------
    begin                : Thi Dec 12 2002
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CButtonGroup_H__
#define __CButtonGroup_H__

#include <FL/Fl_Group.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Input_.H>
#include <FL/Fl_Button.H>
#include <sptk3/CIntList.h>
#include <sptk3/gui/CControl.h>
#include <sptk3/gui/CInput.h>
#include <sptk3/CStrings.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Base class for CCheckButtons and CRadioButtons.
///
/// Implements the most important data communication methods for these button groups.
class SP_EXPORT CButtonGroup : public CControl {
    int         m_maxHeight;        ///< Required group height
protected:
    std::string m_lastValue;        ///< Last value of the group (last list of choices)
    CStrings    m_buttonLabels;     ///< Button labels for the buttons inside
    Fl_Button  *m_otherButton;      ///< The 'Other' Button if requested (add '*' in the button list)
    CInput_    *m_otherInput;       ///< The 'Other' Input if requested (add '*' in the button list)

    /// Internal callback processing
    virtual void controlDataChanged();

    /// Finds a button by label
    /// @param buttonLabel const char *, button label
    /// @returns button index, or -1 if not found
    int  buttonIndex(const char *buttonLabel);

    /// Deselects all buttons
    void deselectAllButtons();

    /// Selects button by index
    void selectButton(int choice);

    /// Creates button. Should be implemented in the derived class.
    virtual Fl_Button *createButton(const char *label,int sz=10,CLayoutAlign layoutAlignment=SP_ALIGN_TOP) = 0;

    /// Constructor initializer
    void ctor_init();

protected:

    /// SPTK-style constructor
    /// @param label const char *, the widget label
    /// @param layoutSize int, the size of widget in layout
    /// @param layoutAlignment CLayoutAlign, widget align in the layout
    CButtonGroup(const char * label=0,int layoutSize=20,CLayoutAlign layoutAlignment=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// FLTK-style constructor
    /// @param x int, widget x-coordinate
    /// @param y int, widget y-coordinate
    /// @param w int, widget width
    /// @param h int, widget height
    /// @param label int, optional widget label
    CButtonGroup(int,int,int,int,const char * = 0);
#endif

public:
    /// @brief Sets the list of the buttons.
    /// @param buttonList CStrings, list of the buttons
    void buttons(const CStrings& buttonList);

    /// @brief Returns the list of the buttons.
    const CStrings& buttons() const { return m_buttonLabels; }

    /// @brief Clears the list of buttons.
    virtual void clearButtons();

    /// @brief Returns the currently selected button(s) as pipe ('|') separated string
    virtual CVariant data() const;

    /// @brief Sets the currently selected button(s)
    ///
    /// Buttons are presented as pipe ('|') separated string.
    /// If the button group allows only one button to be selected at a time
    /// (like radio buttons), only the first item of the string will be used.
    virtual void data(const CVariant v);

    /// @brief Loads the the currently selected button(s)
    ///
    /// Buttons should be presented as pipe ('|') separated string.
    virtual void load(CQuery *);

    /// @brief Saves the the currently selected button(s)
    ///
    /// Buttons are presented as pipe ('|') separated string
    virtual void save(CQuery *);

    /// @brief Loads control data from XML
    ///
    /// Layout information may also include widget size and position,
    /// as well as visible() and active() states
    /// @param node CXmlNode*, the XML node
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be stored
    virtual void     load(const CXmlNode* node,CLayoutXMLmode xmlMode);

    /// @brief Saves control data to XML
    ///
    /// Layout information may also include widget size and position,
    /// as well as visible() and active() states
    /// @param node CXmlNode*, the XML node
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be stored
    virtual void save(CXmlNode* node,CLayoutXMLmode xmlMode) const;

    /// @brief Tells if the the current data content is valid
    ///
    /// Always true for this widget.
    virtual bool valid() const {
        return true;
    }

    /// @brief Computes the preferred size of the button group based on its contents
    /// @param w int&, the optimal width 
    /// @param h int&, the optimal height 
    /// @returns true if the size is stable (doesn't depend on input sizes)
    virtual bool preferredSize(int& w,int& h);
};
/// @}
}
#endif
