/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CColumn.h  -  description
                             -------------------
    begin                : Thu Apr 27 2000
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CCOLUMNS_H__
#define __CCOLUMNS_H__

#include <string>
#include <sptk3/CStrings.h>
#include <sptk3/CVariant.h>
#include <sptk3/cxml>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Class CColumn represents a column descriptor for CListView
class SP_EXPORT CColumn {
protected:
    std::string  m_name;            ///< Column name
    uint32_t     m_width;       ///< Column width in pixels
    CVariantType m_type;            ///< Column data type
    bool         m_visible;     ///< Column visibility
    bool         m_autoWidth;   ///< Column auto width flag
public:

    /// @brief Constructor
    /// @param cname std::string, column name
    /// @param type CVariantType, column data type
    /// @param cwidth int16_t, column width
    /// @param cvisible bool, column visibility
    CColumn(std::string cname="",CVariantType type=VAR_STRING,int16_t cwidth=100,bool cvisible=true);

    /// @brief Copy constructor
    /// @param col const CColumn&, source column object
    CColumn(const CColumn& col);

    /// @brief Returns column name
    std::string name()        const {
        return m_name;
    }

    /// @brief Returns column width in pixels
    int16_t   width()           const {
        return m_width;
    }

    /// @brief Sets column width in pixels
    void    width(int16_t w)          {
        m_width = w;
    }

    /// @brief Returns column visibility
    bool    visible()         const {
        return m_visible;
    }

    /// @brief Sets column visibility
    void    visible(bool v)         {
        m_visible = v;
    }

    /// @brief Returns Column auto width flag
    bool    autoWidth()       const {
        return m_autoWidth;
    }

    /// @brief Sets column auto width flag
    void    autoWidth(bool v)       {
        m_autoWidth = v;
    }

    /// @brief Returns column data type
    CVariantType type()       const {
        return m_type;
    }

    /// @brief Loads column information
    ///
    /// Loads information about column
    /// from the registry.
    /// @param node const CXmlNode&, an XML node to load column information from
    void load(const CXmlNode& node);

    /// @brief Loads column information
    ///
    /// Loads information about column
    /// from the registry.
    /// @param node const CXmlNode*, an XML node to load column information from
    void load(const CXmlNode* node) {
        load(*node);
    }

    /// @brief Saves column information
    ///
    /// Saves information about column
    /// from the registry.
    /// @param node CXmlNode&, an XML node to unload column information to
    void save(CXmlNode& node) const;

    /// @brief Saves column information
    ///
    /// Saves information about column
    /// from the registry.
    /// @param node CXmlNode*, an XML node to unload column information to
    void save(CXmlNode* node) const {
        save(*node);
    }
};

/// Class CColumnList is a list of columns used in CListView
class SP_EXPORT CColumnList : public std::vector<CColumn> {
public:

    /// @brief Default constructor
    CColumnList() {}

    /// @brief Finds the column index by column name
    /// @param columnName const char *, column name
    int  indexOf(const char *columnName) const;

    /// @brief Loads columns information
    ///
    /// Loads information about columns
    /// from the registry.
    /// @param node const CXmlNode&, an XML node to load columns information from
    void load(const CXmlNode& node);

    /// @brief Loads columns information
    ///
    /// Loads information about columns
    /// from the registry.
    /// @param node const CXmlNode*, an XML node to load columns information from
    void load(const CXmlNode* node) {
        load(*node);
    }

    /// @brief Saves columns information
    ///
    /// Saves information about columns
    /// from the registry.
    /// @param node CXmlNode&, an XML node to unload columns information to
    void save(CXmlNode& node) const;

    /// @brief Saves columns information
    ///
    /// Saves information about columns
    /// from the registry.
    /// @param node CXmlNode*, an XML node to unload columns information to
    void save(CXmlNode* node) const {
        save(*node);
    }
};
/// @}
}
#endif
