/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          ccontrol.h  -  description
                             -------------------
    begin                : Tue Apr 4 2000
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CCONTROL_H__
#define __CCONTROL_H__

#include <FL/Fl_Group.H>
#include <FL/Fl_Menu_Button.H>

#include <sptk3/db/CQuery.h>
#include <sptk3/gui/CEvent.h>
#include <sptk3/gui/CLayoutClient.h>
#include <sptk3/cxml>

#include <string>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Control kind is the constant to report the internal SPTK RTTI.
enum CControlKind {
    DCV_UNKNOWN,                 ///< Control kind is unknown
    DCV_BOX=1,                   ///< Control is a box of plain text
    DCV_HTMLBOX=2,               ///< Control is a box of HTML text
    DCV_STRING=4,                ///< Control is a text input (single line)
    DCV_MEMO=8,                  ///< Control is a text input (multiple lines)
    DCV_INTEGER=0x10,            ///< Control is an integer value input
    DCV_FLOAT=0x20,              ///< Control is an integer value input
    DCV_DATE=0x40,               ///< Control is a date input
    DCV_TIME=0x80,               ///< Control is a time input
    DCV_DATETIME=0x100,          ///< Control is a date and time input
    DCV_DATEINTERVAL=0x200,      ///< Control is a date interval input
    DCV_COMBO=0x400,             ///< Control is a combobox
    DCV_INTVALUECOMBO=0x800,     ///< Control is a combobox with int key value
    DCV_LISTBOX=0x1000,          ///< Control is a listbox
    DCV_CHECKBUTTONS=0x2000,     ///< Control is a set of check buttons
    DCV_RADIOBUTTONS=0x4000,     ///< Control is a set of radio buttons
    DCV_PHONE=0x8000,            ///< Control is a phone number input
    DCV_TREEVIEW=0x10000,        ///< Control is a tree widget
    DCV_GROUP=0x20000,           ///< Control is a group
    DCV_SLIDER=0x40000,          ///< Control is a value slider
    DCV_PROGRESS=0x80000         ///< Control is a progress bar
};

/// @brief Special control flags (used as bit combination)
enum FGENTRYFLAGS {
    FGE_NONE,                   ///< No flags
    FGE_SPELLCHECK=1,           ///< Requires spell checking
    FGE_MANDATORY=2,            ///< Must have some value
    FGE_CFIELD=4,               ///< Is a custom field
    FGE_SINGLELINEENTRY=256,    ///< Is a single-line entry
    FGE_MULTILINEENTRY=512      ///< Is a multi-line entry
};

/// @brief Default height for the single-row edit control
#define SINGLE_EDIT_ENTRY_HEIGHT 23

class CDBDropDownList;
class CPackedStrings;
class CColumnList;
class CStrings;

/// @brief Internal field name validation
SP_EXPORT bool checkFieldName(std::string fldName);

/// @brief Base class for most of SPTK widgets.
///
/// It is a group used as a container for the widget(s) inside.
/// Every CControl has a field name, and it can be a layout client.
class SP_EXPORT CControl : public ::Fl_Group, public CLayoutClient {
    friend class CInternalComboBoxPanel;
    friend class CEditor;
    friend class CBaseButton;
    friend class CRadioButton;
    friend class CCheckButton;
    friend class CInput_;

    /// Constructor initializer
    void ctor_init(const char *);

protected:

    bool              m_limited;     ///< Control has the defined min and max values
    uint32_t          m_labelWidth;  ///< Control label width (on the left)
    Fl_Color          m_labelColor;  ///< Control label color
    Fl_Color          m_textColor;   ///< Control text color
    Fl_Font           m_textFont;    ///< Control text font
    uchar             m_textSize;    ///< Control text font size
    std::string       m_fieldName;   ///< Field name for the universal data connection
    std::string       m_hint;        ///< The text of the hint
    Fl_Widget*        m_control;     ///< The main widget inside the control.
    int               m_tag;         ///< Control's special data tag
    /// @brief Control's special flags (a bit combination of FGENTRYFLAGS).
    /// @see FGENTRYFLAGS for details.
    uint32_t          m_controlFlags;
    Fl_Menu_Button   *m_menuButton;  ///< Control's menu button - to show right-mouse click menu
    CEventInfo        m_event;       ///< The last SPTK event inside this control.
    bool              m_hasFocus;    ///< Does the control have focus?

protected:

    /// @brief Does the control contain focus?
    bool containsFocus() const;

    /// @brief Internal focus notification on focus change
    void notifyFocus(bool gotFocus=true);

    /// @brief Internal focus notification on focus change
    virtual void     onEnter();

    /// @brief Internal focus notification on focus change
    virtual void     onExit();

    /// @brief Computes the label height based on the labelFont() and labelWidth()
    uint32_t         labelHeight() const;

    /// @brief Special handle() method
    int              handle(int);

    /// @brief Internal callback function
    static void internalCallback(Fl_Widget *internalWidget,void *data);

public:

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    CControl(const char *label,int layoutSize=20,CLayoutAlign layoutAlign=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CControl(int,int,int,int,const char * = 0);
#endif

public:

    /// @brief Resizes the control and inside widgets.
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    virtual void     resize(int x,int y,int w,int h);

    /// @brief Draws the control
    virtual void     draw();

    /// @brief Returns the control's user data tag
    virtual int      tag() const {
        return m_tag;
    }

    /// @brief Sets the control's user data tag
    virtual void     tag(int t)  {
        m_tag = t;
    }

    /// @brief Returns the control's text font
#if FL_MAJOR_VERSION < 2
    virtual Fl_Font  textFont() const;

    /// @brief Sets the control's text font
    virtual void     textFont(Fl_Font);
#else

    virtual Font    *textFont() const;

    /// @brief Sets the control's text font
    virtual void     textFont(Font *);
#endif

    /// @brief Returns the control's text font size
    virtual uchar    textSize() const;

    /// @brief Sets the control's text font size
    virtual void     textSize(uchar);

    /// @brief Returns the control's text color
    virtual Fl_Color textColor() const;

    /// @brief Sets the control's text color
    virtual void     textColor(Fl_Color);

    /// @brief Returns the control's label color
    virtual Fl_Color labelColor() const;

    /// @brief Sets the control's label color
    virtual void     labelColor(Fl_Color);

    /// @brief Returns the control's background color
    virtual Fl_Color color() const;

    /// @brief Sets the control's background color
    virtual void     color(Fl_Color);

    /// @brief Returns the control's flags
    uint32_t flags() const {
        return m_controlFlags;
    }

    /// @brief Sets the control's flags
    void flags(uint32_t flags);

    /// @brief Returns main widget inside the container
    Fl_Widget*       control() const {
        return m_control;
    }

    /// @brief Returns control's label
    const std::string& label() const {
        return m_label;
    }

    /// @brief Sets control's label
    void             label(const std::string);

    /// @brief Returns control's menu
    Fl_Menu_*        menu() const;

    /// @brief Sets control's menu
    void             menu(const Fl_Menu_Item*);

    /// @brief Returns control's label font
    Fl_Font          labelFont() const {
        return labelfont();
    }

    /// @brief Sets control's label font
    void             labelSize(Fl_Font f) {
        labelfont(f);
    }

    /// @brief Returns control's label font size
    uchar            labelSize() const;

    /// @brief Sets control's label font size
    void             labelSize(uchar);

    /// @brief Returns control's label width
    uint32_t         labelWidth() const {
        return m_labelWidth;
    }

    /// @brief Sets control's label width
    void             labelWidth(uint32_t);

    /// @brief Sets control's hint (tooltip)
    /// @param str std::string, tooltip text
    void             hint(std::string str) {
        if (m_control) {
            m_hint = str;
            m_control->tooltip(m_hint.c_str());
        }
    }

    /// @brief Returns control's hint (tooltip)
    std::string     hint() const {
        return m_hint;
    }

    /// @brief Returns control's max input length, if applicable
    virtual int      maxLength() const {
        return 0;
    }

    /// @brief Sets control's max input length, if applicable
    virtual void     maxLength(int) {}

    /// @brief Returns control's field name for universal data connection
    virtual const std::string& fieldName() const {
        return m_fieldName;
    }

    /// @brief Sets control's field name for universal data connection
    virtual void     fieldName(const std::string&);

    /// @brief Returns control's kind (internal SPTK RTTI).
    virtual CControlKind kind() const = 0;

    /// @brief Returns control's class name (internal SPTK RTTI).
    virtual std::string className() const {
        return "control";
    }

    /// @brief Sets the control to empty text (for the input entries)
    ///
    /// Also sets no selection state / default selection (for the lists)
    virtual void     reset() {
        data("");
    }

    /// @brief Loads control data from query fields
    virtual void     load(CQuery *) {}

    /// @brief Saves control data to query params
    virtual void     save(CQuery *) {}

    /// @brief Loads control data from XML
    ///
    /// Layout information may also include widget size and position,
    /// as well as visible() and active() states
    /// @param node CXmlNode*, the XML node
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be stored
    virtual void     load(const CXmlNode* node,CLayoutXMLmode xmlMode);

    /// @brief Saves control data to XML
    ///
    /// Layout information may also include widget size and position,
    /// as well as visible() and active() states
    /// @param node CXmlNode*, the XML node
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be stored
    virtual void save(CXmlNode* node,CLayoutXMLmode xmlMode) const;

    /// @brief Returns true if the control state is valid, and the data is inside the limits (if applicable)
    virtual bool     valid() const = 0;

    /// @brief The universal data connector
    /// @returns control data
    virtual CVariant data() const {
        return CVariant("");
    }

    /// @brief The universal data connector, sets control data
    virtual void     data(const CVariant v) {}

    /// @brief Fires the event generated by this control
    void             fireEvent(CEvent ev,int32_t arg);

    /// @brief Returns the last event fired by this control
    const CEventInfo& event() const {
        return m_event;
    }

    /// @brief Returns the event type for the last fired by this control
    CEvent           eventType() const {
        return m_event.type();
    }

    /// @brief Returns the event argument for the last fired by this control
    int              eventArgument() const {
        return m_event.argument();
    }

    /// @brief Converts control name to control kind
    static  CControlKind controlNameToType(std::string typeName,int& maxLength,std::string values="");

    /// @brief Control data assignment
    CControl& operator = (const std::string& str) {
        data(str);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (const char *str)    {
        data(str);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (int32_t v)              {
        data(v);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (uint32_t v)         {
        data(v);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (int64_t v)              {
        data(v);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (uint64_t v)         {
        data(v);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (float v)            {
        data(v);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (double v)           {
        data(v);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (CDateTime dt)       {
        data(dt);
        return *this;
    }

    /// @brief Control data assignment
    CControl& operator = (CField& fld)        {
        data(fld);
        return *this;
    }

    /// @brief Control data conversion
    operator std::string () const   {
        return data();
    }

    /// @brief Control data conversion
    operator int32_t () const       {
        return data();
    }

    /// @brief Control data conversion
    operator uint32_t () const  {
        return data();
    }

    /// @brief Control data conversion
    operator int64_t () const       {
        return data();
    }

    /// @brief Control data conversion
    operator uint64_t () const  {
        return data();
    }

    /// @brief Control data conversion
    operator float () const     {
        return data();
    }

    /// @brief Control data conversion
    operator double () const    {
        return data();
    }

    /// @brief Control data conversion
    operator CDateTime () const {
        return data();
    }

public:
    /// @brief Default callback function to support CControl default menu item 'Copy'
    static void defaultControlMenuCopy(Fl_Widget *,void *);

    /// @brief Default callback function to support CControl default menu item 'Delete'
    static void defaultControlMenuCut(Fl_Widget *,void *);

    /// @brief Default callback function to support CControl default menu item 'Paste'
    static void defaultControlMenuPaste(Fl_Widget *,void *);

    /// @brief Default callback function to support CControl default menu item 'Clear'
    static void defaultControlMenuClear(Fl_Widget *,void *);

    /// @brief CControl default menu
    static const Fl_Menu_Item defaultControlMenu[];
};

/// @brief Creates a single control
///
/// The utility function that allows to create a control with one step
/// @param controlKind int, a control kind
/// @param label std::string, a control label
/// @param fieldName std::string, a control field name
/// @param layoutSize int, a control layout size
CControl *createControl(int controlKind,std::string label,std::string fieldName,int layoutSize);

/// @brief Create a control or a group of controls
///
/// Every XML node in xmlControls should be CONTROL node
/// and may contain the following elements:
///   type       a control type, one of the
///      {
///        AREA CODE, CURRENCY, CREDIT CARD, COMBO, CHOICE, CHECK BUTTONS,
///        DATE, DATE AND TIME, FLOAT, GROUP, HTML, INTEGER, LABEL, LONG ZIP CODE,
///        MASK, MULTIPLE CHOICE, MONEY, PHONE, PHONE EXT, PLAIN TEXT,
///        PHONE NUMBER, PHONE EXTENSION, SIMPLE, STRING, SSN, TIME, ZIP CODE
///      }
///   label      a label, optional
///   fieldName  a field name, optional
///   values     a list of values separated with '|' for list-based controls, string, optional
///   sqlValues  an SQL query to build the list values, string, optional, presented as a sub-node with CDATA
///   visible    visibility flag, optional, true or false (1/0)
///   enable     enable flage, optional, true or false (1/0)
///   align      layout align, optional, one of { top,bottom,left,right,client}, the default is top
///   size       layout size in pixels, optional
///
/// The group node may contain other controls. If the control type or parameter is not recognized,
/// the exception is thrown.
/// @param xmlControls const CXmlNodeList&, the controls description in XML
void createControls(const CXmlNodeList& xmlControls) throw(std::exception);

/// @}
}
#endif
