/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CDateTimeInput.h  -  description
                             -------------------
    begin                : Tue Apr 4 2000
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CDATETIMEINPUT_H__
#define __CDATETIMEINPUT_H__

#include <sptk3/sptk.h>

#include <sptk3/CDateTime.h>
#include <sptk3/gui/CInput.h>
#include <sptk3/gui/CControl.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

class CPopupCalendar;
class CDateControl;

/// @brief Base date and time input widget
///
/// Base class for most of the date and time inputs (except fot CDateIntervalInput)
class SP_EXPORT CDateTimeBaseInput : public CInput {
    typedef class CInput inherited;

    /// @brief Mininmum date and time value to check in valid()
    CDateTime            m_minValue;

    /// @brief Maximum date and time value to check in valid()
    CDateTime            m_maxValue;

    /// @brief Constructor initializer
    void ctor_init();

protected:

    /// @brief Popup calendar window
    CPopupCalendar      *m_calendarWindow;

    /// @brief Date input control
    CDateControl        *m_dateInput;

    /// @brief Time input control
    CInput_             *m_timeInput;

    /// @brief Loads data from query
    virtual void load(CQuery *);

    /// @brief Saves data to query
    virtual void save(CQuery *);

    /// @brief Loads control data from XML
    ///
    /// Layout information may also include widget size and position,
    /// as well as visible() and active() states
    /// @param node CXmlNode*, the XML node
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be stored
    virtual void     load(const CXmlNode* node,CLayoutXMLmode xmlMode);

    /// @brief Saves control data to XML
    ///
    /// Layout information may also include widget size and position,
    /// as well as visible() and active() states
    /// @param node CXmlNode*, the XML node
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be stored
    virtual void save(CXmlNode* node,CLayoutXMLmode xmlMode) const;

    /// @brief Returns true if the input data is valid
    virtual bool valid() const;

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    /// @param autoCreate bool, auto create CInput_ widget. Internal.
    CDateTimeBaseInput(const char * label,int layoutSize,CLayoutAlign layoutAlign,bool autoCreate);

    /// @brief Internal callback function
    static void calendarButtonPressed(Fl_Widget *btn,void *data);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CDateTimeBaseInput(int x,int y,int w,int h,const char * label,bool autoCreate);
#endif

    /// @brief Shows drop-down calendar
    bool showCalendar(Fl_Widget *btn);

public:

    /// @brief Sets limits for the value inside
    /// @param limited bool, true if use the limits
    /// @param min CDateTime, minimum value
    /// @param max CDateTime, maximum value
    void setLimits(bool limited,CDateTime min=double(0),CDateTime max=double(0));

    /// @brief Returns widget date or time  value
    virtual CDateTime dateTimeValue() const;

    /// @brief Sets widget date or time value
    virtual void     dateTimeValue(CDateTime dt);
};

class CTimeInput;

/// @brief Date input widget
///
/// Masked date input with the drop-down calendar
class SP_EXPORT CDateInput : public CDateTimeBaseInput {
    /// @brief Constructor initializer
    void ctor_init();
public:

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    CDateInput(const char * label=0,int layoutSize=10,CLayoutAlign layoutAlign=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CDateInput(int x,int y,int w,int h,const char * label=0);
#endif

    ~CDateInput();

    /// @brief Returns the control kind, SPTK-style RTTI
    /// @see CControlKind for more information
    virtual CControlKind kind() const {
        return DCV_DATE;
    }

    /// @brief Returns the control class name, SPTK-style RTTI
    virtual std::string className() const {
        return "date_input";
    }

    /// @brief Universal data connection, returns data from date control.
    /// Data is returned as VAR_DATE.
    virtual CVariant data() const;

    /// @brief Universal data connection, sets data to date control
    virtual void     data(const CVariant v);

    /// @brief Resizes the control and inside widgets.
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    virtual void resize(int x,int y,int w,int h);

    /// Computes the optimal widget height
    /// @param h int&, input - height offered by the program, output - height required by widget
    virtual void preferredHeight(int& h) const;

    /// @brief Creates a widget based on the XML node information
    static CLayoutClient* creator(CXmlNode *node);
};

/// @brief Time input widget
///
/// Masked time input
class SP_EXPORT CTimeInput : public CDateTimeBaseInput {
public:

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    CTimeInput(const char * label=0,int layoutSize=10,CLayoutAlign layoutAlign=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CTimeInput(int x,int y,int w,int h,const char * label=0);
#endif

    /// @brief Returns the control kind, SPTK-style RTTI
    /// @see CControlKind for more information
    virtual CControlKind kind() const {
        return DCV_TIME;
    }

    /// @brief Returns the control class name, SPTK-style RTTI
    virtual std::string className() const {
        return "time_input";
    }

    /// @brief Computes the optimal widget width
    /// @param w int&, input - width offered by the program, output - width required by widget
    virtual void preferredWidth(int& w) const;

    /// @brief Creates a widget based on the XML node information
    static CLayoutClient* creator(CXmlNode *node);
};

/// @brief Date and time input widget.
///
/// Combines date and time input widgets. Includes two input boxes and the drop-down calendar
/// for the date input.
class SP_EXPORT CDateTimeInput : public CDateTimeBaseInput {

    /// @brief Constructor initializer
    void ctor_init();

public:

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    CDateTimeInput(const char * label=0,int layoutSize=10,CLayoutAlign layoutAlign=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CDateTimeInput(int x,int y,int w,int h,const char * label=0);
#endif

    /// @brief Returns the control kind, SPTK-style RTTI
    /// @see CControlKind for more information
    virtual CControlKind kind() const {
        return DCV_DATETIME;
    }

    /// @brief Returns the control class name, SPTK-style RTTI
    virtual std::string className() const {
        return "date_time_input";
    }

    /// @brief Resizes the control and inside widgets.
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    virtual void resize(int x,int y,int w,int h);

    /// @brief Returns widget date or time  value
    virtual CDateTime dateTimeValue() const;

    /// @brief Sets widget date or time value
    virtual void     dateTimeValue(CDateTime dt);

    /// @brief Universal data connection, returns data from control
    virtual CVariant data() const;

    /// @brief Universal data connection, sets data from control
    virtual void     data(const CVariant v);

    /// @brief Creates a widget based on the XML node information
    static CLayoutClient* creator(CXmlNode *node);
};
/// @}
}
#endif
