/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CFont.h  -  description
                             -------------------
    begin                : Fri Jul 28 2006
    copyright            : (C) 2006-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/
#ifndef __CFONT_H__
#define __CFONT_H__

#include <sptk3/sptk.h>
#include <string>
#include <map>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Font information
class CFont {
    std::string m_name;   ///< Font name
    uint32_t    m_size;   ///< Font size
    uint32_t    m_color;  ///< Font color
    uint32_t    m_index;  ///< Font index in fonts collection
    uint32_t    m_attributes; ///< Optional font attributes
public:
    /// @brief Constructor
    /// @param name std::string, full font name
    /// @param size uint32_t, font size
    /// @param color uint32_t, font color
    /// @param index uint32_t, font index in the fonts collection
    /// @param attributes uint32_t, optional font attributes
    CFont(std::string name="arial",uint32_t size=10,uint32_t color=0,uint32_t index=0,uint32_t attributes=0) {
        m_name = name;
        m_size = size;
        m_color = color;
        m_index = index;
        m_attributes = attributes;
    }

    /// @brief Copy constructor
    /// @param font const CFont&, font to copy from
    CFont(const CFont& font) {
        m_name = font.m_name;
        m_size = font.m_size;
        m_color = font.m_color;
        m_index = font.m_index;
        m_attributes = font.m_attributes;
    }

    /// @brief Returns font name
    std::string name() const {
        return m_name;
    }

    /// @brief Returns font size
    uint32_t size() const {
        return m_size;
    }

    /// @brief Returns font color
    uint32_t color() const {
        return m_color;
    }

    /// @brief Returns font index in the fonts collection
    uint32_t index() const {
        return m_index;
    }

    /// @brief Returns font attributes
    uint32_t attributes() const {
        return m_attributes;
    }

    /// @brief Returns unique font id
    ///
    /// Font id combines all the font attributes separated with "|"
    std::string id() const;
};

/// @brief CFontsMap is a map font names to fonts
///
/// The map contains just pointers to the fonts stored in outside list.
class CFontsMap : public std::map<std::string,CFont *> {
public:
    /// @brief Constructor
    CFontsMap() {}

    /// @brief Destructor
    ~CFontsMap() {
        clear();
    }

    /// @brief Clears the fonts
    ///
    /// All the memory allocated for CFont objects is released
    void clear();
};

/// @brief CFontsVector is a map font names to fonts
class CFontsVector {
    std::vector<CFont *>    m_items;  ///< Fonts vector
    CFontsMap               m_index;  ///< Font names map
public:
    /// @brief Constructor
    CFontsVector() {}

    /// @brief Destructor
    ~CFontsVector() {
        clear();
    }

    /// @brief Returns font by index
    CFont* operator[] (uint32_t index) {
        return m_items[index];
    }

    /// @brief Returns font by index
    const CFont* operator[] (uint32_t index) const {
        return m_items[index];
    }

    /// @brief Returns font by its name
    CFont* find(std::string fontName) const;

    /// @brief Appends a font to fonts vector
    /// @param font CFont*, font to append
    void push_back(CFont* font);

    /// @returns the size of the fonts vector
    uint32_t size() const {
        return m_items.size();
    }

    /// @brief Clears the fonts
    ///
    /// All the memory allocated for CFont objects is released
    void clear();
};

/// @}
}

#endif
