/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CPngImage.h  -  description
                             -------------------
    begin                : 3 Sep 2006
    copyright            : (C) 2006-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CPNGIMAGE_H__
#define __CPNGIMAGE_H__

#include <FL/Fl.H>
#include <FL/Fl_Image.H>
#include <sptk3/CBuffer.h>

namespace sptk {
/// @brief PNG image that is initialized from the memory buffer
///
/// This class is used by SPTK themes
class CPngImage : public Fl_RGB_Image {

    /// @brief Internal loader from PNG image buffer
    /// @param buffer const CBuffer&, image data
    void load(const CBuffer& buffer);

public:
    /// @brief The way how the resized parts of image are drawn
    ///
    /// By resized parts I mean everything besides four corners.
    /// Corners are rectangles with the side of corner zone
    enum CPatternDrawMode {
        PDM_UNDEFINED,  ///< Undefined and not drawn, should be defined later.
        PDM_TILE,       ///< The resized parts are tiled
        PDM_STRETCH     ///< The resized parts are stretched
    };

    /// @brief Draws a part of the image by cutting a corresponding source and tiling it in required area
    /// @param srcX int, x-coordinate of the source fragment
    /// @param srcY int, y-coordinate of the source fragment
    /// @param srcW int, width of the source fragment
    /// @param srcH int, height of the source fragment
    /// @param destX int, x-coordinate of the destination fragment
    /// @param destY int, y-coordinate of the destination fragment
    /// @param destW int, width of the destination fragment
    /// @param destH int, height of the destination fragment
    void cutTileDraw(int srcX,int srcY,int srcW,int srcH,int destX,int destY,int destW,int destH);

    /// @brief Draws a part of the image by cutting a corresponding source and stretching it in required area
    /// @param srcX int, x-coordinate of the source fragment
    /// @param srcY int, y-coordinate of the source fragment
    /// @param srcW int, width of the source fragment
    /// @param srcH int, height of the source fragment
    /// @param destX int, x-coordinate of the destination fragment
    /// @param destY int, y-coordinate of the destination fragment
    /// @param destW int, width of the destination fragment
    /// @param destH int, height of the destination fragment
    void cutStretchDraw(int srcX,int srcY,int srcW,int srcH,int destX,int destY,int destW,int destH);

public:
    /// @brief Constructor
    ///
    /// An object of the class contains it's own copy of the image data
    /// @param imagedata const CBuffer&, PNG image data in memory buffer
    CPngImage(const CBuffer& imagedata);

    /// @brief Constructor
    ///
    /// An object of the class contains it's own copy of the image data
    /// @param image Fl_RGB_Image*, RGB image data in memory buffer
    CPngImage(const Fl_RGB_Image* image);

    /// @brief Constructor
    ///
    /// An object of the class contains it's own copy of the image data
    /// @param fileName std::string, image file (.png)
    CPngImage(std::string fileName);

    /// @brief Draws resized image
    ///
    /// @param x int, the x coordinate to draw image
    /// @param y int, the y coordinate to draw image
    /// @param w int, the width to draw image
    /// @param h int, the height to draw image
    /// @param cornerZone int, the height (and width) of the corners that are simply copied from the image w/o processing 
    /// @param drawMode CPatternDrawMode, the mode to draw the resized parts of image
    /// @param drawBackground bool, if true then the internal area of the image is used for background
    void drawResized(int x,int y,int w,int h,int cornerZone,CPatternDrawMode drawMode,bool drawBackground);
};

}

#endif
