/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CScroll.h  -  description
                             -------------------
    begin                : Sun May 22 2003
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CSCROLL_H__
#define __CSCROLL_H__

#include <FL/Fl_Group.h>

#include <sptk3/gui/CScrollBar.h>
#include <sptk3/gui/CLayoutClient.h>
#include <sptk3/gui/CLayoutManager.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Different modes for scrollbars
enum CScrollbarMode { // values for type()
    HORIZONTAL = 1,
    VERTICAL = 2,
    BOTH = 3,
    ALWAYS_ON = 4,
    HORIZONTAL_ALWAYS = 5,
    VERTICAL_ALWAYS = 6,
    BOTH_ALWAYS = 7
};

/// @brief Scroll area widget
///
/// Extended version of FLTK Fl_Group that can be a layout manager and layout client
/// at the same time.
class CScroll : public ::Fl_Group, public CLayoutManager {

    int m_xposition;        ///< Current x-position of the scrolled area
    int m_yposition;        ///< Current y-position of the scrolled area
    int m_width;            ///< Current width of the scrolled area
    int m_height;           ///< Current height of the scrolled area
    int m_oldx;             ///< Former x-position of the scrolled area
    int m_oldy;             ///< Former y-position of the scrolled area

    /// @brief Horizontal scrollbar callback
    static void hscrollbar_cb(Fl_Widget*, void*);

    /// @brief Vertical scrollbar callback
    static void scrollbar_cb(Fl_Widget*, void*);

    /// @brief Constructor initializer
    void ctor_init();

    /// @brief Insure the scrollbars are the last children:
    void fix_scrollbar_order();

    /// @brief Draw the visible area. This is also the callback for fl_scroll.
    static void draw_clip(void* v,int X, int Y, int W, int H);

    /// @brief Computed area for widgets
    void bbox(int& X, int& Y, int& W, int& H);

public:

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    CScroll(const char * label=0,int layoutSize=10,CLayoutAlign layoutAlign=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CScroll(int x,int y,int w,int h,const char *label=0L);
#endif

    CScrollBar scrollbar;     ///< Vertical scrollbar
    CScrollBar hscrollbar;    ///< Horizontal scrollbar

    /// @brief Resizes the group and inside widgets.
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    virtual void resize(int x,int y,int w,int h);

    /// @brief Custom draw() method
    virtual void draw();

    /// @brief Computes the optimal group size
    /// @param w int&, input - width offered by the program, output - width required by widget
    /// @param h int&, input - height offered by the program, output - height required by widget
    /// @returns true if the size is stable (doesn't depend on input sizes)
    virtual bool preferredSize(int& w,int& h);

    /// @brief Removes all the child widgets
    virtual void clear();

    /// @brief Sets the position of the scrolled area
    virtual void position(int X, int Y);

    /// @brief Returns the x-position of the scrolled area
    int xposition() const {
        return m_xposition;
    }

    /// @brief Returns the y-position of the scrolled area
    int yposition() const {
        return m_yposition;
    }

    /// @brief Custom handle method
    int handle(int);

    /// @brief Returns the current label
    std::string label() const {
        return m_label;
    }

    /// @brief Sets the new label
    ///
    /// @param lbl const char*, new label
    void label(const char* lbl) {
        CLayoutClient::label(lbl);
    }

    /// Sets label for the group, makes an internal copy of the string
    /// @param lbl const string&, new label
    void label(const std::string& lbl) {
        CLayoutClient::label(lbl);
    }

    /// @brief Creates a widget based on the XML node information
    /// @param node CXmlNode*, an XML node with widget information
    static CLayoutClient* creator(CXmlNode* node);

    /// @brief Returns widget class name (internal SPTK RTTI).
    virtual std::string className() const {
        return "scroll";
    }
};
/// @}
}
#endif
