/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CThemes.h  -  description
                             -------------------
    begin                : Fri Nov 08 2003
    copyright            : (C) 2003-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CTHEMES_H__
#define __CTHEMES_H__

#include <FL/Fl.h>
#include <sptk3/gui/CFrame.h>
#include <sptk3/gui/CRect.h>
#include <sptk3/gui/CFont.h>
#include <sptk3/gui/CIcon.h>
#include <sptk3/gui/CTabImage.h>
#include <sptk3/cxml>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Class CButtonImage provides a storage for two images (color and desaturated).
class CButtonImage {
public:

    CPngImage *m_colorImage;   ///< Color image
    CPngImage *m_grayImage;    ///< Desaturated image
public:

    /// @brief Constructor
    /// @param image CPngImage *, image to use as color image. The desaturated counterpart is generated automatically.
    CButtonImage(CPngImage *image=0L);

    /// Destructor
    ~CButtonImage();

    /// @brief Deallocates the allocated resources (if any) and sets images to NULL
    void clear();

    /// @brief Sets an image to use
    /// @param image CPngImage *, image to use as color image. The desaturated counterpart is generated automatically.
    /// @param desaturateInactive bool, true if you need to desaturate button image when mouse is not over it
    void set(CPngImage *image,bool desaturateInactive);
};

/// @brief Theme button type - defines which routine to use for which FLTK button type
enum CThemeButtonType { THM_BUTTON_UP, THM_BUTTON_DOWN, THM_DEFAULT_BUTTON_UP, THM_DEFAULT_BUTTON_DOWN };

/// @brief Theme button size - defines the button size and look to select the correct pixmap
enum CThemeButtonSize { THM_BUTTON_COMBO, THM_BUTTON_CHECK, THM_BUTTON_RADIO, THM_BUTTON_NORMAL, THM_BUTTON_LARGE };

/// @brief Theme scroll bar type - defines the scroll bar direction
enum CThemeScrollBarType { THM_SCROLLBAR_VERTICAL, THM_SCROLLBAR_HORIZONTAL };

/// @brief The complete set of four images to draw a scroll bar
class CScrollBarImages {
public:

    /// @brief Type of the scroll bar image
    enum type {
        NORMAL,  ///< Image of the scroll bar without slider
        ACTIVE,  ///< Image of the scroll bar without slider, with buttons pushed down
        SLIDER,  ///< Image of the scroll bar slider
        MARK     ///< Image of the scroll bar slider mark (in the center)
    };

    CPngImage     *m_images[4];    ///< Images of scrollbar elements
    int            m_buttonSize;   ///< Width of the button in horizontal scrollbar
public:
    /// @brief Default constructor
    CScrollBarImages() {
        memset(m_images,0,sizeof(m_images));
        m_buttonSize = 0;
    }

    /// @brief Loads the images from the theme files
    /// @param btnSize int, the size of the button in theme
    /// @param atype CThemeScrollBarType, scroll bar type
    void load(int btnSize,CThemeScrollBarType atype);

    /// @brief Clears the scrollbar images for the default theme
    void clear();

    /// @brief Returns the button size - width of the button in horizontal scrollbar
    int buttonSize() {
        return m_buttonSize;
    }

    /// @brief Returns image by index
    /// @param ndx CScrollBarImages::type, a type (index) of image
    CPngImage* operator[] (type ndx) { return m_images[ndx]; }
};

/// @brief Widget themes
///
/// Class CThemes is a static loader for the themes. Only one one instance of the class should exist, and it's created automatically.
/// The class allows to load a set of routines to draw the particular elements of the interface, and the images for these routines.
class CThemes {
    friend class CBaseButton;

    static std::string m_name;              ///< Theme name

    static std::map<std::string,Fl_Color> m_colorNames; ///< Map of a symbolic names to Fl_Colors

    static Fl_Box_Draw_F *m_flatBox;        ///< Original flat box draw function
    static Fl_Box_Draw_F *m_thinUpBox;      ///< Original thin up box draw function
    static Fl_Box_Draw_F *m_upBox;          ///< Original up box draw function
    static Fl_Box_Draw_F *m_thinDownBox;    ///< Original thin down box draw function
    static Fl_Box_Draw_F *m_downBox;        ///< Original down box draw function

    static Fl_Box_Draw_F *m_thinUpFrame;    ///< Original thin up frame draw function
    static Fl_Box_Draw_F *m_upFrame;        ///< Original up frame draw function
    static Fl_Box_Draw_F *m_thinDownFrame;  ///< Original thin down frame draw function
    static Fl_Box_Draw_F *m_downFrame;      ///< Original down frame draw function

    static CFontsMap      m_fonts;          ///< Available fonts
    static CFrames        m_frames;         ///< Available frames

    /// An array of the background images to replace the original FLTK colors
    /// FL_DARK3..FL_BACKGROUND..FL_LIGHT3
    static CPngImage     *m_background[7];

    /// Four different types of button pixmaps - for normal button, normal pressed button, default button, and default pressed button accordingly.
    static CButtonImage   m_normalButtons[4];

    /// Two different types of combo box button pixmaps - for normal button, and normal pressed button. Two extra array members are reserved.
    static CButtonImage   m_comboButtons[4];

    /// Two different types of check box button pixmaps - for check box unchecked, and check box checked. Two extra array members are reserved.
    static CButtonImage   m_checkButtons[4];

    /// Two different types of radio button pixmaps - for radio button unchecked, and radio button checked. Two extra array members are reserved.
    static CButtonImage   m_radioButtons[4];

    /// @brief Two different types of large button pixmaps - for large button, and large pressed button. Two extra array members are reserved.
    static CButtonImage   m_largeButtons[4];

    /// @brief Two images for painting a progress bar
    ///
    /// One image is for the frame, one for the progress bar itself. Images should have the
    /// compatible size. If you put the progress bar image on top its frame - they should show
    /// a picture of 100% progress
    static CPngImage      *m_progressBar[2];

    /// @brief Three images for painting tabs
    ///
    /// One image is for the inactive tab, one for the active tab, and one for the frame on the bottom of tabs
    static CTabImages      m_tabImages;

    /// @brief Images for painting a scroll bar (horizontal and vertical)
    ///
    /// Images should have the compatible size.
    /// If you put the slider image on top its frame - they should show
    /// a picture of 100% progress
    static CScrollBarImages m_scrollBar[2];

    /// Flag indicating if buttons should be drawn as flat bars if mouse isn't over the button
    static bool           m_flatButtons;

    /// Theme registry. Contains several importang theme settings.
    static CXmlDoc       *m_registry;

    /// Restores the original FLTK theme
    static void reset();

    /// Returns the correct image array for the given button size
    /// @param sz CThemeButtonSize, the general type of the button. @see CThemeButtonSize for possible button types.
    static CButtonImage *sizeToButtonImages(CThemeButtonSize sz);

    /// Loads and sets the pixmap from the file for the given button size and type
    /// @param sz CThemeButtonSize, theme button size
    /// @param bt CThemeButtonType, theme button type
    /// @param fileName std::string, pixmap file name
    static void replaceButton(CThemeButtonSize sz,CThemeButtonType bt,std::string fileName);

    /// Loads and sets the pixmap from the file for the given pixmap array member
    /// @param ndx int, image index
    /// @param fileName std::string, pixmap file name
    static void replaceImage(CPngImage *images[],int ndx,std::string fileName);

    /// Paints the rectangle with the background image.
    /// @param xx int, rectangle x-coordinate
    /// @param yy int, rectangle y-coordinate
    /// @param ww int, rectangle width
    /// @param hh int, rectangle height
    /// @param image CPngImage *, image to use
    static void paintBackground(int xx,int yy,int ww,int hh,CPngImage *image);

    /// @brief Draws theme box with optional frame
    /// @param x int, box x-coordinate
    /// @param y int, box y-coordinate
    /// @param w int, box width
    /// @param h int, box height
    /// @param color Fl_Color, box color
    /// @param frameType Fl_Boxtype, box frame type
    static void drawThemeBox(int x,int y,int w,int h,Fl_Color color,Fl_Boxtype frameType=FL_NO_BOX);

    /// @brief Draws theme box with optional frame
    /// @param x int, box x-coordinate
    /// @param y int, box y-coordinate
    /// @param w int, box width
    /// @param h int, box height
    /// @param frameType Fl_Boxtype, box frame type
    static void drawThemeFrame(int x,int y,int w,int h,Fl_Boxtype frameType);

    /// Determines the background image (if any) for the given color in range
    /// FL_DARK3..FL_LIGHT3
    static CPngImage *imageForColor(Fl_Color clr);

    /// Flag the shows if theme buttons should be desaturated for the inactive
    /// state of the button
    static bool m_desaturateInactiveButtons;

    /// Defines a radius of the focus frame' corner
    static int  m_buttonFocusRadius;

    /// Draws a check or radio button image
    ///
    /// @param buttonImages CButtonImage*, an array of four button images
    /// @param x int, x-coordinate of the button
    /// @param y int, y-coordinate of the button
    /// @param checked bool, is button checked?
    /// @param highlited bool, is button highlited?
    /// @returns false, if theme has no pixmap for the button, or true otherwise
    static bool drawCheckOrRadioButton(CButtonImage* buttonImages,int x,int y,bool checked,bool highlited);

protected:
    static CIconMap m_icons[4];   ///< Icons for use with buttons (four different sizes available, @see CIconSize)

public:

    /// Default constructor
    CThemes();

    /// Destructor
    ~CThemes();

    /// Loads and sets the theme. The themeName is the directory with the theme images.
    static void set(std::string themeName);

    /// @brief Returns a button icon by symbolic name
    /// @param iconName string, symbolic name for an icons
    /// @param iconSize CIconSize, the size of the icon
    /// @returns pointer to the icon, or NULL if not found
    static CIcon* getIcon(std::string iconName,CIconSize iconSize);

    /// @brief Returns a button icon image by symbolic name
    /// @param iconName string, symbolic name for an icons
    /// @param iconSize CIconSize, the size of the icon
    /// @returns pointer to the image, or NULL if not found
    static CPngImage* getIconImage(std::string iconName,CIconSize iconSize);

    /// @brief Registers a new button icon
    /// @param icon CIcon*, pointer to the icon
    /// @param iconSize CIconSize, the size of the icon
    static void registerIcon(CIcon* icon,CIconSize iconSize);

    /// The replacement rountine to draw  FL_FLAT_BOX
    static void drawFlatBox(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_THIN_UP_BOX
    static void drawThinUpBox(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_UP_BOX
    static void drawUpBox(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_THIN_DOWN_BOX
    static void drawThinDownBox(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_DOWN_BOX
    static void drawDownBox(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_THIN_UP_FRAME
    static void drawThinUpFrame(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_UP_FRAME
    static void drawUpFrame(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_THIN_DOWN_FRAME
    static void drawThinDownFrame(int x,int y,int w,int h,Fl_Color clr);

    /// The replacement rountine to draw  FL_DOWN_FRAME
    static void drawDownFrame(int x,int y,int w,int h,Fl_Color clr);

    /// The rountine to draw theme button
    /// @param sz CThemeButtonSize,
    /// @param x int, button x-coordinate
    /// @param y int, button y-coordinate
    /// @param w int&, button width, may be altered by button to meet the button size requirements
    /// @param h int&, button height, may be altered by button to meet the button size requirements
    /// @param highlited bool, true to draw the highlighted button
    /// @param pressed bool, true to draw pressed button
    /// @param defaultButton bool, true to draw default button
    static bool drawButton(CThemeButtonSize sz,int x,int y,int& w,int& h,bool highlited,bool pressed,bool defaultButton);

    /// Computes and returns the correct size of the button in the parameters.
    /// @param sz CThemeButtonSize, the general type of the button. @see CThemeButtonSize for possible button types.
    /// @param w int&, button width, may be altered by button to meet the button size requirements
    /// @param h int&, button height, may be altered by button to meet the button size requirements
    /// @returns false, if theme has no pixmap for the button, or true otherwise
    static bool sizeButton(CThemeButtonSize sz,int& w,int& h);

    /// Computes and returns the minimum size of the progress bar
    /// @param w int&, progress bar width, may be altered by progress bar to meet the progress bar size requirements
    /// @param h int&, progress bar height, may be altered by progress bar to meet the progress bar size requirements
    /// @returns false, if theme has no pixmap for the progress bar, or true otherwise
    static bool sizeProgressBar(int& w,int& h);

    /// Computes and returns the minimum size of the scroll bar
    /// @param w int&, scroll bar width, may be altered by scroll bar to meet the scroll bar size requirements
    /// @param h int&, scroll bar height, may be altered by scroll bar to meet the scroll bar size requirements
    /// @returns false, if theme has no pixmaps for the scroll bar, or true otherwise
    static bool sizeScrollBar(int& w,int& h);

    /// Returns the button focuse corner radius
    static int  buttonFocusRadius() {
        return m_buttonFocusRadius;
    }

    /// Draws a check button image
    /// @returns false, if theme has no pixmap for the button, or true otherwise
    /// @param x int, x-coordinate of the button
    /// @param y int, y-coordinate of the button
    /// @param checked bool, is button checked?
    /// @param highlited bool, is button highlited?
    static bool drawCheckButton(int x,int y,bool checked,bool highlited);

    /// Draws a radio button image
    /// @returns false, if theme has no pixmap for the button, or true otherwise
    /// @param x int, x-coordinate of the button
    /// @param y int, y-coordinate of the button
    /// @param checked bool, is button checked?
    /// @param highlited bool, is button highlited?
    static bool drawRadioButton(int x,int y,bool checked,bool highlited);

    /// Draws a progress bar image
    /// The height of the image is defined by the 'progress0.png' image of the theme.
    /// The width of the image is adjusted as required, but it can't be less than
    /// returned by sizeProgressBar() method.
    /// @returns false, if theme has no pixmaps for the progress bar, or true otherwise
    /// @param x int, x-coordinate of the progress bar
    /// @param y int, y-coordinate of the progress bar
    /// @param w int, width of the progress bar
    /// @param percent float, percent value of the progress bar
    static bool drawProgressBar(int x,int y,int w,float percent);

    /// @brief Draws a scroll bar image.
    ///
    /// The type of the scroll bar is defined by max of (w,h).
    /// @returns false, if theme has no pixmaps for the scroll bar, or true otherwise
    /// @param x int, x-coordinate of the scroll bar
    /// @param y int, y-coordinate of the scroll bar
    /// @param w int, width of the scroll bar
    /// @param h int, width of the scroll bar
    /// @param minimum float, minimum value of the scroll bar
    /// @param maximum float, maximum value of the scroll bar
    /// @param value float, value of the scroll bar
    /// @param size float, size of the scroll bar
    /// @param activeZone int, 0 - no active zone, 1 - button 1, 2 - button 2, 3 - slider, +8 - pushed
    /// @param active bool, true if scrollbar is active (enabled)
    static bool drawScrollBar(int x,int y,int w,int h,float minimum,float maximum,float value,float size,int activeZone,bool active);

    /// @brief Draws a tab.
    /// @returns false, if theme has no pixmaps for the tab, or true otherwise
    /// @param x int, x-coordinate of the scroll bar
    /// @param y int, y-coordinate of the scroll bar
    /// @param w int, width of the scroll bar
    /// @param h int, width of the scroll bar
    /// @param active bool, true if stab is active
    static bool drawTab(int x,int y,int w,int h,bool active);

    /// @brief Draws a tab frame
    ///
    /// Draws a bottom frame for the tabs
    /// @returns false, if theme has no pixmaps for the tab frame, or true otherwise
    /// @param x int, x-coordinate of the scroll bar
    /// @param y int, y-coordinate of the scroll bar
    /// @param w int, width of the scroll bar
    /// @param h int, width of the scroll bar
    static bool drawTabFrame(int x,int y,int w,int h);

    /// Returns the list of the available themes
    static CStrings availableThemes();

    /// @brief Themes root directories
    ///
    /// On different system it may be a different set of directories.
    /// The first available correct combination of theme directory and theme name
    /// defines the theme directory
    static const CStrings& searchDirectories();

    /// @brief All the available screen fonts
    static CFontsVector   screenFonts;

    /// @brief Returns theme name
    static std::string name() { return m_name; }

    /// @brief Theme tar archive. Contains all the theme data
    static CTar           m_tar;
};

/// @}
}
#endif
