/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CTreeView.h  -  description
                             -------------------
    begin                : June 21 2003
    copyright            : (C) 2002-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CTREECONTROL_H__
#define __CTREECONTROL_H__

#include <FL/Fl_Widget.h>
#include <FL/Fl_Pixmap.h>
#include <sptk3/gui/CLayoutClient.h>
#include <sptk3/gui/CScroll.h>
#include <sptk3/gui/CGroup.h>

#include <string>
#include <vector>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

class CTreeItem;
class CTreeControl;

/// @brief A callback that allows to create a widget of required type and use it as an item body
///
/// An item body should be any CLayoutClient-derived widget.
/// By default (if that callback isn't redefined, SPTK creates CBox widget.
/// @param item CTreeItem*, a tree item that would be a parent for the new item
typedef CLayoutClient* (*CTreeItemCreator)(CTreeItem *item);

/// @brief Tree widget item.
///
/// A group widget with the extra information about item pixmaps
/// and current state.
class CTreeItem : public CGroup {
    int             m_itemHeight;   ///< Returns the height of the item text
    int             m_indent;       ///< The width of the indent zone
    int             m_labelWidth;   ///< The width of the label computed by fl_measure
    int             m_labelHeight;  ///< The height of the label computed by fl_measure
    bool            m_selected;     ///< Is the item selected (controlled by parent tree control)
    Fl_Color        m_itemColor[2]; ///< Item colors (text,bg) before selection

protected:
    CLayoutClient  *m_body;         ///< The widget that represents the item without a child
    const Fl_Image *m_openedImage;  ///< The image for the open state
    const Fl_Image *m_closedImage;  ///< The image for the closed state
    bool            m_opened;       ///< The item's state
    CTreeControl   *m_tree;         ///< The tree control

    /** Adds a child item to the item using the path. The required path items are created automatically.
     * This methos is used internally only.
     * @param pathFolders const std::vector<std::string>&, the path to the new child item relatively to the item
     * @param offset uint32_t, the offset in pathFolders where the path starts 
     * @param openedImage Fl_Image, the image for the opened folder
     * @param closedImage Fl_Image, the image for the closed folder
     * @param itemImage Fl_Image, the image for the child item
     * @param data void *, the user data or ID attached to the item
     * @returns the new child item
     */
    CTreeItem *addPathOffset(const std::vector<std::string>& pathFolders,uint32_t offset,const Fl_Image *openedImage,const Fl_Image *closedImage,const Fl_Image *itemImage=0L,void *data=0L);

public:
    static const Fl_Image* treeOpened;            ///< Default image of the opened tree
    static const Fl_Image* treeClosed;            ///< Default image of the closed tree
    static const Fl_Image* folderOpened;          ///< Default image of the opened floder
    static const Fl_Image* folderClosed;          ///< Default image of the closed floder
    static const Fl_Image* document;              ///< Default image of the document

public:
    /** Constructor. If the closedImage parameter is omitted the openedImage is used instead.
     * @param label const char *, the item label
     * @param openedImage Fl_Image, the image for the opened state
     * @param closedImage Fl_Image, the image for the closed state
     * @param data void *, the user data or ID attached to the item
     */
    CTreeItem(const char *label,const Fl_Image *openedImage=0L,const Fl_Image *closedImage=0L,void *data=0L);

    /** Adds a child item to the item. If the closedImage parameter is omitted the openedImage is used instead.
     * @param label const char *, the item label
     * @param openedImage Fl_Image, the image for the opened state
     * @param closedImage Fl_Image, the image for the closed state
     * @param data void *, the user data or ID attached to the item
     * @returns the new child item
     */
    CTreeItem *addItem(const char *label,const Fl_Image *openedImage=0L,const Fl_Image *closedImage=0L,void *data=0L);

    /** Adds a child item to the item using the path. The required path items are created automatically.
     * @param path const std::vector<std::string>&, the path to the new child item relatively to the item
     * @param openedImage Fl_Image, the image for the opened folder
     * @param closedImage Fl_Image, the image for the closed folder
     * @param itemImage Fl_Image, the image for the child item
     * @param data void *, the user data or ID attached to the item
     * @returns the new child item
     */
    CTreeItem *addPath(const std::vector<std::string>& path,const Fl_Image *openedImage,const Fl_Image *closedImage,const Fl_Image *itemImage=0L,void *data=0L);

    /** Adds a child item to the item using the path. The required path items are created automatically.
     * The default images are used for the folders in the path.
     * @param path const std::vector<std::string>&, the path to the new child item relatively to the item
     * @param itemImage Fl_Image, the image for the child item
     * @param data void *, the user data or ID attached to the item
     * @returns the new child item
     */
    CTreeItem *addPath(const std::vector<std::string>& path,const Fl_Image *itemImage=0L,void *data=0L);

    /// The widget that represents the item without children
    CLayoutClient *body() {
        return m_body;
    }

    /// Sets the item's caption
    /// @param lbl const char *, the new item caption
    void label(const char *lbl);

    /// Returns the item's caption
    const char * label() const {
        return m_body->label().c_str();
    }

    /// Removes all the child items of the item
    virtual void clear();

    /// Finds an item among the child items
    /// @returns the child item (if found) or NULL
    CTreeItem *findItem(const char *label) const;

    /// Finds an item by the item user data in the whole tree
    /// @param data void *, user data to find
    /// @returns the child item (if found) or NULL
    CTreeItem *findData(const void *data) const;

    /// Removes an item from the child items.
    /// Only the intermediate children are considered
    /// @param item CTreeItem *, item to remove
    void removeItem(CTreeItem *item);

    /// Returns the parent item
    /// @returns the parent item or NULL (for the root item)
    CTreeItem *parentItem() const;

    /// Returns the tree control this item belongs to
    CTreeControl *tree() const {
        return m_tree;
    }

    /// Moves the immediate child before another immediate child
    /// If beforeItem is 0L then moves to the end of the children
    /// @param item CTreeItem *, the item to move
    /// @param beforeItem CTreeItem *, the item to insert before
    void moveItem(CTreeItem *item,CTreeItem *beforeItem=0L);

    /// Moves the immediate child up or down in the list of items
    /// @param item CTreeItem *, the item to move
    /// @param direction int, the direction to move. -1 to move up, 1 to move down
    void moveItem(CTreeItem *item,int direction);

    /// Shows or hides all the child items. Changes the item state accordingly.
    /// @param vc bool, show(true) or hide(false) children
    void visibleChildren(bool vc);

    /// Shows all the child items. Changes the item state accordingly.
    void open()         {
        visibleChildren(true);
    }

    /// Hides all the child items. Changes the item state accordingly.
    void close()        {
        visibleChildren(false);
    }

    /// Reports the item state - opened or closed.
    bool opened() const {
        return m_opened;
    }

    /// Reports the item indent from the left.
    uint32_t indent() const {
        return m_indent;
    }
    /// Sets image(s) for the item
    /// @param openedImage Fl_Image *, the image for the opened state
    /// @param closedImage Fl_Image *, the image for the closed state
    void setImage(const Fl_Image *openedImage,const Fl_Image *closedImage=0L) {
        m_openedImage = openedImage;
        m_closedImage = closedImage;
    }

    /// The event handle function. Internal. See Fl_Widget for details.
    /// @param event int, event type.
    /// @returns true if event was processed.
    int handle(int event);

    /// The draw function. Internal. See Fl_Widget for details.
    virtual void draw();

    /// Computes the preferred size of the item based on the font of the parent widget,
    /// the image size, and the text (label) of the item.
    /// @param w int, input/output desirable widget width
    /// @param h int, input/output desirable widget heigth
    /// @returns true if the size is stable (doesn't depend on input sizes)
    virtual bool preferredSize(int& w,int& h);

    /// Resizes item and sub items
    void resize(int xx,int yy,int ww,int hh);

    /// @brief Selects or unselects item
    /// @param flag bool, true is item shoud be selected
    virtual void select(bool flag);

    /// @brief Returns selection state of the item
    bool selected() const { return m_selected; }

    /// @brief Selects the next item in the tree
    bool selectNext();

    /// @brief Selects the prior item in the tree
    bool selectPrior();

    /// @brief Selects the last item in the tree
    bool selectFirst();

    /// @brief Selects the first item in the tree
    bool selectLast();

    /// @brief Selects the first visible item in the child tree
    CTreeItem* findFirst() const;

    /// @brief Selects the last visible item in the child tree
    /// @param recursive bool, should we search the child items
    CTreeItem* findLast(bool recursive) const;

    /// @brief Selects the next visible item in the child tree
    /// @param recursive bool, should we search the child items
    CTreeItem* findNext(bool recursive) const;

    /// @brief Selects the prior visible item in the child tree
    /// @param recursive bool, should we search the child items
    CTreeItem* findPrior(bool recursive) const;

    /// @brief Returns widget class name (internal SPTK RTTI).
    virtual std::string className() const {
        return "treeitem";
    }
};

/// @brief Vector of CTreeItem pointers
typedef std::vector<CTreeItem*> CTreeItemVector;

/// @brief Tree widget.
///
/// Designed to be used inside
/// CTreeView widget, but also can be used by itself. It's missing for data connection
/// support of CTreeView, though.
class CTreeControl : public CScroll {
    friend class CTreeItem;
private:
    CTreeItemVector   m_selectedItems; ///< The selected items
    CTreeItem*        m_root;          ///< The root item
protected:
    bool              m_tabPressed;    ///< Internal flag to temporarely block tree items focus acceptance
    CTreeItemCreator  m_itemCreator;   ///< The tree item creator creates a tree item' body, @see CTreeItemCreator

    /// @brief The default tree item creator
    ///
    /// Creates a tree item' body as CBox, @see CTreeItemCreator
    static CLayoutClient* defaultItemCreator(CTreeItem *item);
public:
    /// The constructor.
    /// @param label const char *, the widget's label
    /// @param layoutSize int, size of widget in layout. See CLayoutClient for details
    /// @param align CLayoutAlign, widget align in the layout
    CTreeControl(const char *label,int layoutSize=50,CLayoutAlign align=SP_ALIGN_TOP);

    /// Adds a child item to the root item. If the closedImage parameter is omitted the openedImage is used instead.
    /// @param label const char *, the item label
    /// @param openedImage Fl_Image, the image for the opened state
    /// @param closedImage Fl_Image, the image for the closed state
    /// @param data void *, the user data or ID attached to the item
    /// @returns the new child item
    CTreeItem *addItem(const char *label,const Fl_Image *openedImage=0L,const Fl_Image *closedImage=0L,void *data=0L);

    /// Adds a child item to the root item using the path. The required path items are created automatically.
    /// Path elements are separated with '/'. The default images are used for the folders in the path.
    /// @param path const std::vector<std::string>&, the path to the new child item
    /// @param openedImage Fl_Image, the image for the folders in opened state
    /// @param closedImage Fl_Image, the image for the folders in closed state
    /// @param itemImage Fl_Image, the image for the child item
    /// @param data void *, the user data or ID attached to the item
    /// @returns the new child item
    CTreeItem *addPath(const std::vector<std::string>& path,const Fl_Image *openedImage,const Fl_Image *closedImage,const Fl_Image *itemImage=0L,void *data=0L);

    /** Adds a child item to the root item using the path. The required path items are created automatically.
     * Path elements are separated with '/'. The default images are used for the folders in the path.
     * @param path const std::vector<std::string>&, the path to the new child item
     * @param itemImage Fl_Image, the image for the child item
     * @param data void *, the user data or ID attached to the item
     * @returns the new child item
     */
    CTreeItem *addPath(const std::vector<std::string>& path,const Fl_Image *itemImage=0L,void *data=0L);

    /// @brief Finds an item by the item label in the whole tree
    /// @param item const char *, item label to find
    /// @returns the child item (if found) or NULL
    CTreeItem *findItem(const char *item) const;

    /// @brief Finds an item by the item user data in the whole tree
    /// @param data void *, user data to find
    /// @returns the child item (if found) or NULL
    CTreeItem *findData(const void *data) const;

    /// @brief Removes all the items in the tree.
    virtual void clear() {
        m_selectedItems.clear();
        m_root->clear();
    }

    /// @brief Moves the item in the tree.
    /// @param item CTreeItem *, item to insert
    /// @param beforeItem CTreeItem *, item before the insert point. If 0L - it's moved to the very last position in parent item list of items.
    void moveItem(CTreeItem *item,CTreeItem *beforeItem=0L);

    /// @brief Removes all the item and underlying structure from the tree.
    /// @param item CTreeItem *, item to remove
    void removeItem(CTreeItem *item);

    /// @brief Reports the root item.
    CTreeItem *root() const {
        return m_root;
    }

    /// @brief Reports the currently selected item.
    CTreeItem *selected() const {
        if (m_selectedItems.empty())
            return 0;
        return m_selectedItems[0];
    }

    /// @brief Selects only one item.
    /// @param item CTreeItem, item to select
    /// @param giveFocus bool, should the item be focused?
    void selectOnly(CTreeItem *item,bool giveFocus=false);

    /// @brief Makes the item visible if it's outside the visible area.
    /// @param item CTreeItem, item to make visible
    void makeVisible(CTreeItem *item);

    /// @brief Special handle() function.
    /// @param event int, an FLTK event
    /// @returns true, if event was processed
    int handle(int event);

    /// @brief The tree item creator creates a tree item' body, @see CTreeItemCreator
    void itemCreator(CTreeItemCreator ic) {
        m_itemCreator = ic;
    }

    /// @brief Loads group controls data from XML node
    ///
    /// @param node const CXmlNode&, node to load data from
    /// @param autoCreate bool, create widgets if they are not found
    virtual void load(const CXmlNode& node,bool autoCreate=false) throw(CException);

    /// @brief Loads group controls data from XML node
    ///
    /// @param node const CXmlNode*, node to load data from
    /// @param autoCreate bool, create widgets if they are not found
    virtual void load(const CXmlNode* node,bool autoCreate=false)  throw(CException) {
        load(*node,autoCreate);
    }

    /// @brief Saves group controls data into XML node
    ///
    /// @param node const CXmlNode&, node to save data into
    virtual void save(CXmlNode& node) const;

    /// @brief Saves group controls data into XML node
    ///
    /// @param node const CXmlNode*, node to save data into
    virtual void save(CXmlNode* node) const {
        save(*node);
    }

    /// @brief Returns widget class name (internal SPTK RTTI).
    virtual std::string className() const {
        return "tree";
    }
};
/// @}
}
#endif
