/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CTreeView.h  -  description
                             -------------------
    begin                : June 21 2003
    copyright            : (C) 2002-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CTREEVIEW_H__
#define __CTREEVIEW_H__

#include <sptk3/gui/CPngImage.h>
#include <sptk3/gui/CTreeControl.h>
#include <sptk3/gui/CDataControl.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Tree view widget.
///
/// Shows tree items with icons. Can use different icons for opened/closed folders.
class SP_EXPORT CTreeView : public CControl {

    /// @brief The actual tree control
    CTreeControl *m_treeControl;

    /// @brief Constructor initializer
    void ctor_init();
protected:

    /// @brief Internal tree control callback
    static void internal_callback(Fl_Widget *,void *);

public:

    /// @brief Constructor in SPTK style
    /// @param label const char *, label
    /// @param layoutSize int, widget align in layout
    /// @param layoutAlign CLayoutAlign, widget align in layout
    CTreeView(const char * label=0,int layoutSize=10,CLayoutAlign layoutAlign=SP_ALIGN_TOP);

#ifdef __COMPATIBILITY_MODE__
    /// @brief Constructor in FLTK style
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    /// @param label, const char * label
    CTreeView(int x,int y,int w,int h,const char *label=0);
#endif

    /// @brief Destructor
    ~CTreeView();

    /// @brief Returns controls' kind (internal SPTK RTTI).
    virtual CControlKind kind() const {
        return DCV_TREEVIEW;
    }

    /// @brief Returns controls' class name (internal SPTK RTTI).
    virtual std::string className() const {
        return "CTreeView";
    }

    /// @brief Adds a child item to the item. If the closedImage parameter is omitted the openedImage is used instead.
    /// @param label std::string, the item label
    /// @param openedImage const Fl_Image *, the image for the opened state
    /// @param closedImage const Fl_Image *, the image for the closed state
    /// @param data void *, the user data or ID attached to the item
    /// @returns the new child item
    CTreeItem *addItem(std::string label,const Fl_Image *openedImage,const Fl_Image *closedImage=0L,void *data=0L) {
        return m_treeControl->addItem(label.c_str(),openedImage,closedImage,data);
    }

    /// @brief Adds a child item to the item using the path. The required path items are created automatically.
    /// @param path const CStrings&, the path to the new child item relatively to the item
    /// @param openedImage const Fl_Image *, the image for the opened folder
    /// @param closedImage const Fl_Image *, the image for the closed folder
    /// @param itemImage const Fl_Image *, the image for the child item
    /// @param data void *, the user data or ID attached to the item
    /// @returns the new child item
    CTreeItem *addPath(const std::vector<std::string>& path,const Fl_Image *openedImage,const Fl_Image *closedImage,const Fl_Image *itemImage=0L,void *data=0L) {
        return m_treeControl->addPath(path,openedImage,closedImage,itemImage,data);
    }

    /// @brief Adds a child item to the item using the default folder images.
    /// @param path const CStrings&, the item full path in the tree starting with '/'
    /// @param itemImage const Fl_Image *, the image for the child item
    /// @param data void *, the user data or ID attached to the item
    /// @returns the new child item
    CTreeItem *addPath(const std::vector<std::string>& path,const Fl_Image *itemImage=0L,void *data=0L) {
        return addPath(path,itemImage,data);
    }

    /// @brief Removes an item from the parent tree item
    void removeItem(CTreeItem *item) {
        m_treeControl->removeItem(item);
    }

    /// @brief Returns currently selected item
    CTreeItem *selected() const {
        return m_treeControl->selected();
    }

    /// @brief Returns currently selected item path in the tree
    std::string selectedPath() const;

    /// @brief Removes all the tree items
    virtual void clear() {
        m_treeControl->clear();
    }

    /// @brief Relayouts the tree. May be necessary after you've changed items data a lot
    void relayout() {
        m_treeControl->relayout();
        redraw();
    }

    /// @brief Resizes the control and inside widgets.
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    virtual void     resize(int x,int y,int w,int h);

    /// @brief Returns the currently selected item ID (or user_data)
    virtual CVariant data() const;

    /// @brief Selects the item with matching ID (or user_data)
    virtual void     data(const CVariant v);

    /// @brief Data connection isn't implemented yet
    virtual void load(CQuery *);

    /// @brief Data connection isn't implemented yet
    virtual void save(CQuery *);

    /// @brief Loads the the widget from XML node
    ///
    /// The widget information may include widget attributes
    /// and widget data
    /// @param node CXmlNode*, XML node
    virtual void load(const CXmlNode *node);

    /// @brief Saves the the widget to XML node
    ///
    /// The widget information may include widget attributes
    /// and widget data
    /// @param node CXmlNode*, XML node
    virtual void save(CXmlNode *node) const;

    /// @brief Returns tru if data is valid
    virtual bool valid() const {
        return true;
    }

    /// @brief Creates a widget based on the XML node information
    /// @param node CXmlNode*, an XML node with widget information
    static CLayoutClient* creator(CXmlNode* node);
};
/// @}
}
#endif
