/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CWindow.h  -  description
                             -------------------
    begin                : Sun May 22 2003
    copyright            : (C) 2000-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CWINDOW_H__
#define __CWINDOW_H__

#include <FL/Fl_Double_Window.H>
#include <sptk3/gui/CLayoutManager.h>
#include <sptk3/CStrings.h>
#include <sptk3/gui/CWindowShape.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Window widget
///
/// Extended version of FLTK's standard Fl_Double_Window, with added
/// CLayoutManager capabilities
class CWindow : public Fl_Double_Window, public CLayoutManager, public CWindowShape {
public:

    /// @brief Constructor
    /// @param w int, window width
    /// @param h int, window height
    /// @param label int, window label
    CWindow(int w,int h,const char * label=0L)
    : Fl_Double_Window(w,h,label), CLayoutManager(this,10,SP_ALIGN_NONE), CWindowShape(this) {}

    /// @brief Constructor
    /// @param x int, window x-position
    /// @param y int, window y-position
    /// @param w int, window width
    /// @param h int, window height
    /// @param label int, window label
    CWindow(int x,int y,int w,int h,const char *label=0L)
    : Fl_Double_Window(x,y,w,h), CLayoutManager(this,10,SP_ALIGN_NONE), CWindowShape(this) {}

    /// @brief Draws a window, including an optional background image
    virtual void draw();

    /// @brief Custom show method
    virtual void show() {
        Fl_Double_Window::show();
    }

    /// @brief Custom show method
    /// @param argc int, number of parameters in argv[]
    /// @param argv char*[], an array of program command line arguments
    virtual void show(int argc,char* argv[]) {
        Fl_Double_Window::show(argc,argv);
    }

    /// @brief Custom hide method
    virtual void hide();

    /// @brief Relayouts window's widgets that have CLayoutClient interface
    virtual void relayout() {
        resize(x(),y(),w(),h());
    }

    /// @brief Removes all the widgets inside the window
    virtual void clear() {
        Fl_Double_Window::clear();
    }

    /// @brief Resizes the window and inside widgets.
    /// @param x int, x-position
    /// @param y int, y-position
    /// @param w int, width
    /// @param h int, height
    virtual void resize(int x,int y,int w,int h);

    /// @brief Computes the optimal window size
    /// @param w int&, input - width offered by the program, output - width required by widget
    /// @param h int&, input - height offered by the program, output - height required by widget
    /// @returns true if the size is stable (doesn't depend on input sizes)
    virtual bool preferredSize(int& w,int& h);

    /// @brief Custom window events handle
    ///
    /// Mostly used for supporting non-rectangular shape windows.
    int handle(int event);

    /// @brief Loads window coordinates and widgets from XML node
    ///
    /// @param node const CXmlNode*, node to load data from
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be loaded
    virtual void load(const CXmlNode* node,CLayoutXMLmode xmlMode) throw(CException);

    /// @brief Saves window coordinates and widgets into XML node
    ///
    /// @param node CXmlNode*, node to save data into
    /// @param xmlMode CLayoutXMLmode, the mode defining how the layout and/or data should be loaded
    virtual void save(CXmlNode* node,CLayoutXMLmode xmlMode) const;

    /// @brief Loads the window position from XML node
    ///
    /// @param node const CXmlNode&, node to load position from
    void loadPosition(const CXmlNode* node);

    /// @brief Saves the window position into XML node
    ///
    /// @param node CXmlNode&, node to save position into
    void savePosition(CXmlNode* node) const;

    /// @brief Returns the current label
    std::string label() const {
        return m_label;
    }

    /// @brief Sets the new label
    ///
    /// @param lbl const char*, new label
    void label(const char* lbl) {
        CLayoutClient::label(lbl);
    }

    /// Sets label for the group, makes an internal copy of the string
    /// @param lbl const string&, new label
    void label(const std::string& lbl) {
        CLayoutClient::label(lbl);
    }

    /// @brief Returns widget class name (internal SPTK RTTI).
    virtual std::string className() const {
        return "window";
    }
};
/// @}
}
#endif
