/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CWindowShape.h  -  description
                             -------------------
    begin                : Oct 19 2006
    copyright            : (C) 2006-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CWINDOWSHAPE_H__
#define __CWINDOWSHAPE_H__

#include <vector>

#include <sptk3/gui/CShapePoint.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

class CWindow;

/// @brief Window non-rectangular shape support
///
/// Window shape is defined with the array of points that should be modified at every resize of the window.
/// The derived window should modify the shape points array by ovewriting shapeResize() method.
/// The window resize support is activated only if window resizable() is true.
class CWindowShape {
protected:
	sptk::CWindow*           m_window;           ///< Window to change shape
    int                      m_resizingZone;     ///< Resizing mouse zone where mouse was pushed
    int                      m_lastW;            ///< Last window width during the shape generation
    int                      m_lastH;            ///< Last window height during the shape generation
    int                      m_pushedX;          ///< X coordinate of the last FL_PUSH event
    int                      m_pushedY;          ///< Y coordinate of the last FL_PUSH event
    std::vector<CShapePoint> m_shapePoints;      ///< Vector of points defining window shape
    bool                     m_shapeChanged;     ///< Is the shape size changed?
    int                      m_borderWidth;      ///< Shape border width, the default is 6
    bool                     m_borderCleared;    ///< Is no border flag set for the window?
	bool					 m_shapeExtension;	 ///< Is the shape extension activated?

    /// @brief Computes mouse zone from window coordinates
    /// @param mouseX int, mouse x-coordinate in the window
    /// @param mouseY int, mouse y-coordinate in the window
    int  mouseZone(int mouseX,int mouseY) const;

    /// @brief Resizes the window based on mouse zone and mouse screen coordinates
    /// @param mouseX int, mouse x-coordinate in the screen
    /// @param mouseY int, mouse y-coordinate in the screen
    void changeSize(int mouseX,int mouseY);

    /// @brief Changes the window cursor based on the mouse position in the window
    ///
    /// Designed to emulate cursor changes on the window border for shaped windows.
    /// Should be uncoditionally called inside handle() for the window.
    virtual int  shapeCursorHandle(int event);

public:
    /// @brief Constructor
    /// @param window sptk::CWindow*, window to change shape
	CWindowShape(sptk::CWindow* window);

    /// @brief Destructor
    virtual ~CWindowShape() {}

    /// @brief Returns true if the window is resizing as a result of border drag
    ///
    /// Can be used to simplified drawing during the resize
    bool shapeResizing() {
        return m_resizingZone != 0;
    }

protected:

	/// @brief Initializes shape extension
	///
	/// Allows window to use shape extension. Should be called
	/// before the window is first shown.
	void initShapeExtension();

	/// @brief Returns true if border was cleared for the shape
	bool borderCleared() const { return m_borderCleared; }

    /// @brief Sets the shape to the window during window->draw()
    void shapeApply();

    /// @brief Calls resizeShape() from inside CWindow::resize()
    void resizingShape(int w,int h) {
        if (m_lastW == w && m_lastH == h)
            return;
        shapeResize(w,h);
        m_shapeChanged = true;
        m_lastW = w;
        m_lastH = h;
    }

    /// @brief Shape resize
    ///
    /// This method should be overwritten in the derived class
    /// to adapt m_shapePoints[] to the window size.
	virtual void shapeResize(int w,int h) {}
};
/// @}
}

#endif
