/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CXmlDocType.h  -  description
                             -------------------
    begin                : June 27 2006
    based on the code    : Mikko Lahteenmaki <Laza@Flashmail.com>
    copyright            : (C) 2006-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CXMLDOCTYPE_H__
#define __CXMLDOCTYPE_H__

#include <sptk3/xml/CXmlEntities.h>

#include <string>
#include <map>

namespace sptk {

/// @addtogroup XML
/// @{

/// @brief XML document type.
///
/// Represents tag <DOCTYPE ...> in XML document.
/// It can return a map of all entities().
/// Provides the name(), public_id() and system_id() functions.
class CXmlDocType {
    friend class CXmlDoc;
    char         m_replacementBuffer[16];   ///< The buffer used to return replacement literals
	CBuffer		 m_encodeBuffers[2];		///< Encode buffers
public:

    /// @brief Default constructor
    CXmlDocType() {}

    /// @brief Constructor
    CXmlDocType(const char *name, const char *public_id=0, const char *system_id=0);

    /// @brief Returns the name of the document type as specified in the <!DOCTYPE name> tag
    const std::string &name() const {
        return m_name;
    }

    /// @brief Returns the public identifier of the external DTD subset
    ///
    /// Returns empty string if there is no public identifier
    const std::string &publicID() const {
        return m_public_id;
    }

    /// @brief Returns the system identifier of the external DTD subset.
    ///
    /// Returns empty string if there is no system identifier
    const std::string &systemID() const {
        return m_system_id;
    }

    /// @brief Returns a map of all entities described in the DTD
    ///
    /// NOTE: Map doesn't hold default entities.
    CXmlEntities &entities() {
        return m_entities;
    }

    /// @brief Returns a map of all entities described in the DTD
    ///
    /// NOTE: Map doesn't hold default entities.
    const CXmlEntities &entities() const {
        return m_entities;
    }

    /// @brief Encodes string to XML representation.
    ///
    /// Converts "<test>" to "&lt;test&gt;"
    /// @returns true, any entities replaced.
    /// @param str const char *, string to convert
    /// @param ret CBuffer&, converted text is stored here
    bool encodeEntities(const char *str, CBuffer& ret);

    /// @brief Decodes entities in string to their actual values.
    ///
    /// Converts "&lt;test&gt;" to "<test>"
    /// @param str const char*, text to convert
    /// @param sz uint32_t, text length
    /// @param ret CBuffer&, converted text is stored here
    void decodeEntities(const char *str, uint32_t sz, CBuffer& ret);

    /// @brief Searches for entity with given name
    ///
    /// @returns true, if attribute is found.
    /// @param name const char *, entity to search
    bool hasEntity(const char *name);

    /// @brief Removes named entity from entity map
    ///
    /// @returns true, if entity removed.
    /// @param name entity to remove
    void removeEntity(const char *name) {
        m_entities.removeEntity(name);
    }

    /// @brief Returnes replacement value for named entity.
    ///
    /// If entity is not found, empty string is returned.
    /// @param name const char *, entity name
    /// @param replacementLength uint32_t&, the length of the replacement 
    const char* getReplacement(const char *name,uint32_t& replacementLength);

    /// @brief Adds an entity to the map
    ///
    /// If entity named 'name' exists already in map,
    /// its value is replaced with 'replacement'
    /// @param name entity to add/change
    /// @param replacement value that represents entity
    void setEntity(const char *name, const char *replacement) {
        m_entities.setEntity(name, replacement);
    }

private:
    CXmlEntities m_entities;    ///< List of entities

    std::string m_name;         ///< Document type name
    std::string m_public_id;    ///< Public ID
    std::string m_system_id;    ///< System ID
};
/// @}
}
#endif
