/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CXmlElement.h  -  description
                             -------------------
    begin                : Sun May 22 2003
    based on the code    : Mikko Lahteenmaki <Laza@Flashmail.com>
    copyright            : (C) 2003-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CXMLELEMENT_H__
#define __CXMLELEMENT_H__

#include <sptk3/CBuffer.h>
#include <sptk3/CStrings.h>
#include <sptk3/CDateTime.h>
#include <sptk3/xml/CXmlValue.h>
#include <sptk3/xml/CXmlAttributes.h>

#include <string>
#include <map>
#include <vector>

namespace sptk {

/// @addtogroup XML
/// @{

class CXmlDoc;

/// @brief XML Element is a named item that may optionally have sub-nodes and attributes
class CXmlElement : public CXmlNamedItem {
    friend class CXmlDoc;
    CXmlNodeList       m_nodes;              ///< The list of subnodes

protected:
    CXmlAttributes     m_attributes;         ///< Node attributes

    /// @brief Protected constructor for creating CXmlDoc only
    ///
    /// @param doc CXmlDoc&, a document.
    CXmlElement(CXmlDoc& doc) : CXmlNamedItem(doc), m_attributes(this) {}

public:
    /// @brief Constructor
    ///
    /// @param parent CXmlNode*, a parent node.
    /// @param tagname const char*, a name of XML tag
    CXmlElement(CXmlNode& parent,const char* tagname)
            : CXmlNamedItem(parent,tagname), m_attributes(this) {}

    /// @brief Constructor
    ///
    /// @param parent CXmlNode*, a parent node.
    /// @param tagname const char*, a name of XML tag
    CXmlElement(CXmlNode* parent,const char* tagname)
            : CXmlNamedItem(*parent,tagname), m_attributes(this) {}

    /// @brief Constructor
    ///
    /// @param parent CXmlNode &, a parent node.
    /// @param tagname const string&, a name of XML tag
    CXmlElement(CXmlNode& parent,const std::string& tagname)
            : CXmlNamedItem(parent,tagname), m_attributes(this) {}

    /// @brief Returns node type
    virtual CXmlNodeType type() const {
        return DOM_ELEMENT;
    }

    /// @brief Returns the first subnode iterator
    virtual iterator begin() {
        return m_nodes.begin();
    }

    /// @brief Returns the first subnode const iterator
    virtual const_iterator begin() const {
        return m_nodes.begin();
    }

    /// @brief Returns the end subnode iterator
    virtual iterator end() {
        return m_nodes.end();
    }

    /// @brief Returns the end subnode const iterator
    virtual const_iterator end() const {
        return m_nodes.end();
    }

    /// @brief Returns a number of subnodes
    virtual uint32_t size() const {
        return (uint32_t) m_nodes.size();
    }

    /// @brief Returns true if node has no subnodes of subnodes
    virtual bool empty() const {
        return m_nodes.empty();
    }

    /// @brief Appends a subnode
    ///
    /// @param node CXmlNode*, node to append
    virtual void push_back(CXmlNode* node);

    /// @brief Inserts a subnode
    ///
    /// @param pos iterator, insert position with the list of subnodes
    /// @param node CXmlNode*, node to insert
    virtual void insert(iterator pos,CXmlNode* node);

    /// @brief Removes a subnode
    ///
    /// Any memory allocated for subnode is released and subnode is
    /// removed from its parent
    /// @param node CXmlNode*, node to remove
    virtual void remove(CXmlNode* node);

    /// @brief Removes a subnode
    ///
    /// Disconnects subnode from parent (this node)
    virtual void unlink(CXmlNode* node);

    /// @brief Deletes all children and clears all the attributes
    ///
    /// Any memory, associated with children or attributes,
    /// is released.
    virtual void clear();

    /// @brief Returns referrence to node attributes
    virtual CXmlAttributes& attributes() {
        return m_attributes;
    }

    /// @brief Returns referrence to node attributes (const version)
    virtual const CXmlAttributes& attributes() const {
        return m_attributes;
    }

    /// Returns true, if node has any attributes
    virtual bool hasAttributes() const {
        return m_attributes.size() != 0;
    }

    /// Returns true, if given attribute is found
    /// @param attr const char *, attribute to search
    virtual bool hasAttribute(const char *attr) const {
        return m_attributes.hasAttribute(attr);
    }

    /// Returns attribute value for given attribute.
    /// HTML tags can have empty attributes, for those you should use has_attribute() method.
    /// @param attr std::string, name of attribute
    /// @param defaultValue const char *, a default value. If attribute doesn't exist then default value is returned.
    /// @returns attribute value
    virtual CXmlValue getAttribute(std::string attr,const char *defaultValue="") const {
        return m_attributes.getAttribute(attr,defaultValue);
    }

    /// Sets new value to attribute 'attr'.
    /// If attribute is not found, it's added to map.
    /// @param attr const char*, attribute name
    /// @param value CXmlValue, attribute value
    /// @param defaultValue const char *, a default value. If attribute value is matching default value than attribute isn't stored (or removed if it existed).
    virtual void setAttribute(const char *attr, CXmlValue value,const char *defaultValue="") {
        m_attributes.setAttribute(attr,value,defaultValue);
    }

    /// Sets new value to attribute 'attr'.
    /// If attribute is not found, it's added to map.
    /// @param attr const string&, an attribute name
    /// @param value CXmlValue, attribute value
    /// @param defaultValue const char *, a default value. If attribute value is matching default value than attribute isn't stored (or removed if it existed).
    virtual void setAttribute(const std::string& attr, CXmlValue value,const char *defaultValue="") {
        m_attributes.setAttribute(attr.c_str(),value,defaultValue);
    }
};

/// @}
}
#endif
