/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CWindowShape.cpp  -  description
                             -------------------
    begin                : Oct 19 2006
    copyright            : (C) 2006-2007 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#include <sptk3/gui/CWindow.h>
#include <FL/fl_draw.h>
#include <FL/x.H>
#include <FL/Fl.H>

using namespace std;
using namespace sptk;

#ifndef _WIN32
#include <X11/extensions/shape.h>
#endif

#include <iostream>

#define TOP_EDGE    1
#define BOTTOM_EDGE 2
#define LEFT_EDGE   4
#define RIGHT_EDGE  8

CWindowShape::CWindowShape(CWindow* window) {
    m_window = window;
    m_shapeChanged = true;
    m_lastW = m_lastH = 0;
    m_resizingZone = 0;
    m_borderWidth = 6;
    m_borderCleared = false;
    m_shapeExtension = false;
}

void CWindowShape::initShapeExtension() {
    m_shapeExtension = true;
#ifdef _WIN32
    if (m_window)
        m_window->clear_border();
#endif
}

void CWindowShape::shapeApply() {
    if (!m_shapeExtension)
        return;
    if (m_shapePoints.size()) {
        CShapePoint* array = &m_shapePoints[0];
#ifdef _WIN32
        HRGN region = CreatePolygonRgn((CONST POINT*)array, m_shapePoints.size(), WINDING);
        SetWindowRgn(fl_xid(m_window), region, TRUE);
        if (!m_borderCleared) {
            //ShowWindow(fl_xid(m_window),SW_HIDE);
            LONG style = GetWindowLong(fl_xid(m_window), GWL_EXSTYLE);
            style |= WS_EX_APPWINDOW;
            SetWindowLong(fl_xid(m_window), GWL_EXSTYLE, style);
            m_borderCleared = true;
            //ShowWindow(fl_xid(m_window),SW_SHOW);
        }
#else
        Region region = XPolygonRegion((XPoint*)array, m_shapePoints.size(), WindingRule);
        //XUnmapWindow(fl_display, fl_xid(m_window));
        XShapeCombineRegion(fl_display, fl_xid(m_window), ShapeBounding, 0, 0, region, ShapeSet);

        if (!m_borderCleared) {
            int xpos = m_window->x();
            int ypos = m_window->y();
            // Defining window with no caption
            long prop[5] = { 3,1,0,0,0 };
            Atom motif_wm_hints = XInternAtom(fl_display, "_MOTIF_WM_HINTS", 0);
            XChangeProperty (fl_display, fl_xid(m_window), motif_wm_hints, motif_wm_hints, 32,
                             PropModeReplace, (unsigned char*) prop, 5);
            m_borderCleared = true;
            //XDeleteProperty (fl_display, fl_xid(m_window),net_wm_state);
            XMoveWindow(fl_display,fl_xid(m_window),xpos,ypos);
        }

        XMapWindow(fl_display, fl_xid(m_window));
        if (region != None)
            XDestroyRegion(region);
#endif

    }
    m_shapeChanged = false;
}

int CWindowShape::mouseZone(int mouseX,int mouseY) const {
    int zone = 0;
    int zoneWidth = m_borderWidth * 2;

    for (unsigned i = 0; i < 2; i++) {
        if (mouseX < zoneWidth) {
            zone = LEFT_EDGE;
        } else
            if (mouseX > m_window->w() - zoneWidth)
                zone = RIGHT_EDGE;
        if (mouseY < m_borderWidth)
            zone |= TOP_EDGE;
        else
            if (mouseY > m_window->h() - zoneWidth)
                zone |= BOTTOM_EDGE;
        if ( zone & (LEFT_EDGE|RIGHT_EDGE) && zone & (TOP_EDGE|BOTTOM_EDGE) )
            return zone; // Mouse is on the corner
        zoneWidth = m_borderWidth;
    }
    return zone;
}

void CWindowShape::changeSize(int mouseX,int mouseY) {
    if (!m_shapeExtension)
        return;
    if (!m_resizingZone)
        return;
    int oldX=m_window->x(), oldY=m_window->y(), oldW=m_window->w(), oldH=m_window->h();
    int newX=oldX, newY=oldY, newW=oldW, newH=oldH;

    if (m_resizingZone & TOP_EDGE) {
        if (oldY != mouseY) {
            newH = oldY + oldH- mouseY;
            newY = mouseY;
        }
    } else
        if (m_resizingZone & BOTTOM_EDGE) {
            if (oldY + oldH != mouseY)
                newH = mouseY - oldY;
        }

    if (m_resizingZone & LEFT_EDGE) {
        if (oldX != mouseX) {
            newX = mouseX;
            newW = oldX + oldW - mouseX;
        }
    } else
        if (m_resizingZone & RIGHT_EDGE) {
            if (oldX + oldW != mouseX)
                newW = mouseX - oldX;
        }

    if (newX != oldX || newY != oldY)
        m_window->resize(newX,newY,newW,newH);
    else
        if (newW != oldW || newH != oldH)
            m_window->size(newW,newH);
}

int CWindowShape::shapeCursorHandle(int event) {
    if (!m_shapeExtension)
        return 0;
    switch (event) {
    case FL_ENTER:
    case FL_LEAVE:
        return 1;
    case FL_MOVE:
        switch (mouseZone(Fl::event_x(),Fl::event_y())) {
        case LEFT_EDGE|TOP_EDGE:
        case RIGHT_EDGE|BOTTOM_EDGE:
            m_window->cursor(FL_CURSOR_NWSE);
            return 1;
        case TOP_EDGE:
        case BOTTOM_EDGE:
            m_window->cursor(FL_CURSOR_NS);
            return 1;
        case LEFT_EDGE:
        case RIGHT_EDGE:
            m_window->cursor(FL_CURSOR_WE);
            return 1;
        case LEFT_EDGE|BOTTOM_EDGE:
        case RIGHT_EDGE|TOP_EDGE:
            m_window->cursor(FL_CURSOR_NESW);
            return 1;
        }
        m_window->cursor(FL_CURSOR_DEFAULT);
        return 1;
    }
    return 0;
}

