/***************************************************************************
                              CDirectoryDS.cpp  -  description
                              -------------------
        begin                : Nov 28 2002
        copyright            : (C) 2002-2007 by Alexey Parshin. All rights reserved.
        email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#include <string>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <ctype.h>

//#include <FL/filename.H>

#include <sptk3/CDirectoryDS.h>
#include <sptk3/CException.h>
#include <sptk3/CSmallPixmapIDs.h>
#include <sptk3/CStrings.h>

#include <sptk3/filedefs.h>

#define CASE_INSENSITIVE 1

#ifdef _WIN32
const char slash = '\\';
#else
const char slash = '/';
#endif

#ifndef FL_ALIGN_LEFT
#define FL_ALIGN_LEFT 4
#endif

using namespace std;
using namespace sptk;

/* That function is adapted from Rich Salz. */
bool fl_file_match(const char *s, const char *p) {
    int nesting;

    for (;;) {
        switch(*p++) {

        case '?' :           // match any single character
            if (!*s++)
                return false;
            break;

        case '*' :           // match 0-n of any characters
            // do trailing * quickly
            if (!*p)
                return true;
            while (!fl_file_match(s, p))
                if (!*s++)
                    return false;
            return true;

        case '[':            // match one character in set of form [abc-d] or [^a-b]
            {
                if (!*s)
                    return false;
                bool reverse = (*p=='^' || *p=='!');
                if (reverse)
                    p++;
                bool matched = false;
                char last = 0;
                while (*p) {
                    if (*p=='-' && last) {
                        if (*s <= *++p && *s >= last )
                            matched = true;
                        last = 0;
                    } else {
                        if (*s == *p)
                            matched = true;
                    }
                    last = *p++;
                    if (*p==']')
                        break;
                }
                if (matched == reverse)
                    return false;
                s++;
                p++;
            }
            break;

        case '{' :           // {pattern1|pattern2|pattern3}
NEXTCASE:
            if (fl_file_match(s,p))
                return true;
            for (nesting = 0;;) {
                switch (*p++) {
                case '\\':
                    if (*p)
                        p++;
                    break;
                case '{':
                    nesting++;
                    break;
                case '}':
                    if (!nesting--)
                        return false;
                    break;
                case '|':
                case ',':
                    if (nesting==0)
                        goto NEXTCASE;
                case 0:
                    return false;
                }
            }
        case '|':            // skip rest of |pattern|pattern} when called recursively
        case ',':
            for (nesting = 0; *p && nesting >= 0;) {
                switch (*p++) {
                case '\\':
                    if (*p)
                        p++;
                    break;
                case '{':
                    nesting++;
                    break;
                case '}':
                    nesting--;
                    break;
                }
            }
            break;
        case '}':
            break;

        case 0:              // end of pattern
            return !*s;

#if CASE_INSENSITIVE

        case '\\':           // quote next character
            if (*p)
                p++;
            if (*s++ != *(p-1))
                return false;
            break;
        default:
            if (*s != *(p-1) && tolower(*s) != *(p-1))
                return false;
            s++;
#else

        case '\\':           // quote next character
            if (*p)
                p++;
        default  :
            if (*s++ != *(p-1))
                return false;
            break;
#endif

        }
    }
}

// Returns typename
string CDirectoryDS::getFileType(const struct stat &st, CSmallPixmapType& image,const char *fname) const {
    bool executable = S_ISEXEC(st.st_mode);
    bool directory = false;
    image = SXPM_DOCUMENT;

    string modeName;
    if (S_ISDIR(st.st_mode)) {
        modeName = "Directory";
        executable = false;
        directory = true;
        image = SXPM_FOLDER;
    } else if (S_ISREG(st.st_mode)) {
        if (executable)
            modeName = "Executable";
        else
            modeName = "File";
    }

    if (executable) {
        image = SXPM_EXECUTABLE;
    } else {
        if (!directory) {
            const char *ext = strrchr(fname,'.');
            const char *sep = strrchr(fname,slash);
            if (ext && ext > sep) {
                ext++;
                if (strcasecmp(ext,"doc") == 0)
                    image = SXPM_DOC_DOCUMENT;
                if (strcasecmp(ext,"txt") == 0)
                    image = SXPM_TXT_DOCUMENT;
                if (strcasecmp(ext,"xls") == 0)
                    image = SXPM_XLS_DOCUMENT;
                if (strcasecmp(ext,"csv") == 0)
                    image = SXPM_XLS_DOCUMENT;
            }
        }
    }

    return modeName;
}

// Define access mode constants if they aren't already defined.
#ifndef R_OK
# define R_OK 04
#endif

// dataset navigation

string absolutePath(string path) {
    char slashStr[] = { slash, 0 };
    char currentDir[256];
    string fullPath;
    getcwd(currentDir,255);
#ifdef _WIN32

    path = replaceAll(path,"/","\\");
    if (path[1] == ':')
        fullPath = path;
    else if (path[0] == '\\') {
        fullPath = string(currentDir).substr(0,2) + path;
    } else
        fullPath = string(currentDir) + slashStr + path;
#else

    path = replaceAll(path,"\\","/");
    if (path[0] == slash)
        fullPath = path;
    else
        fullPath = string(currentDir) + slashStr + path;
#endif

    CStrings pathItems(fullPath,slashStr);
    for (unsigned i = 0; i < pathItems.size(); i++) {
        if ( pathItems[i] == ".." ) {
            pathItems.remove(i);
            i--;
            if (i > 0) {
                pathItems.remove(i);
                i--;
            }
        }
        if ( pathItems[i] == "." ) {
            pathItems.remove(i);
            i--;
        }
    }
    path = pathItems.asString(slashStr);
    if (!path.length())
        path = slashStr;
    return path;
}

void CDirectoryDS::directory(string d) {
    m_directory = absolutePath(d);
}

// read the directory() and move item into the first entry
bool CDirectoryDS::open() {
    clear();

    size_t dlen = m_directory.length()-1;
    if (dlen) {
        if (m_directory[dlen] != slash)
            m_directory += slash;
    }
#ifdef _WIN32
    WIN32_FIND_DATA FindFileData;
    HANDLE hFind = FindFirstFile((m_directory+"*.*").c_str(), &FindFileData);
    if (hFind == INVALID_HANDLE_VALUE)
        return false;
#else

    dirent **files;
    //int num_files = fl_filename_list(m_directory.c_str(), &files);
    int num_files = scandir(m_directory.c_str(), &files, 0, alphasort);
    if (num_files <= 0)
        return false;
#endif

    struct stat st;

    vector<CFieldList*> fileList;
    int n = 0;
    unsigned index = 0;
    do {

#ifdef _WIN32
        char* file = FindFileData.cFileName;
#else

        if (!files[n]) {
            n++;
            continue;
        }
        char* file = (char *)files[n]->d_name;
#endif

        size_t len = strlen(file);
        if (len && file[len-1] == '/')
            file[len-1] = 0;

        if ((showPolicy() & DDS_HIDE_DOT_FILES) && file[0]=='.') {
            n++;
            continue;
        }

        bool is_link = false;
        bool is_dir = false;

        string fullName = m_directory + file;
        lstat(fullName.c_str(), &st);

#ifndef _WIN32

        if ((st.st_mode & S_IFLNK) == S_IFLNK) {
            is_link = true;
            stat(fullName.c_str(), &st);
        }
#endif
        if (!S_ISDIR(st.st_mode)) {
            size_t patternCount = m_pattern.size();
            if (patternCount) {
                bool matchFound = false;
                for (unsigned i = 0; i < patternCount; i++) {
                    if (fl_file_match(file, m_pattern[i].c_str())) {
                        matchFound = true;
                        break;
                    }
                }
                if (!matchFound) {
                    n++;
                    continue;
                }
            }
        } else
            is_dir = true;

        CSmallPixmapType pixmapType;
        string modeName = getFileType(st, pixmapType, file);

        if (is_link) {
            modeName += ' ';
            modeName += "link";
        }

        bool useEntry;
        if (is_dir)
            useEntry = (showPolicy() & DDS_HIDE_DIRECTORIES) == 0;
        else
            useEntry = (showPolicy() & DDS_HIDE_FILES) == 0;

        if (useEntry) {
            CFieldList   *df = new CFieldList(false);
            df->push_back(" ",false).setImageNdx(pixmapType);
            df->push_back("Name",false)      = file;
            if (modeName == "Directory")
                df->push_back("Size",false)   = "";
            else
                df->push_back("Size",false) = (uint32_t)st.st_size;
            df->push_back("Type",false)      = modeName;
            df->push_back("Modified",false)  = CDateTime::convertCTime(st.st_mtime);
            df->push_back("",false)          = (uint32_t)index; // Fake key value
            index++;

            if (access(fullName.c_str(), R_OK)!=0) {
                (*df)[uint32_t(0)].flags = FL_ALIGN_LEFT;
                (*df)[uint32_t(1)].flags = FL_ALIGN_LEFT;
            }

            if (is_dir)
                m_list.push_back(df);
            else
                fileList.push_back(df);
        }

#ifndef _WIN32
        free((struct dirent*)files[n]);
        files[n] = 0;
#endif

        n++;
    }
#ifdef _WIN32
    while (FindNextFile(hFind, &FindFileData));
#else

    while (n < num_files)
        ;
#endif

    for (unsigned i = 0; i < fileList.size(); i++)
        m_list.push_back( fileList[i] );
    fileList.clear();

    first();

    int defaultWidths[5] = { 3, 30, 10, 10, 16 };
    if (m_current)
        for (unsigned f = 0; f < 5; f++) {
            (*this)[f].flags = FL_ALIGN_LEFT;
            (*this)[f].width = defaultWidths[f];
        }

#ifndef _WIN32
    free((struct dirent**)files);
#endif

    return m_list.size()>0;
}
