/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CRWLock.cpp  -  description
                             -------------------
    begin                : Thu Apr 28 2005
    copyright            : (C) 2005-2007 by Ilya A. Volynets-Evenbakh
    email                : ilya@total-knowledge.com
    based on             : CWaiter.cpp by Alexey Parshin. All rights reserved.
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#include <sptk3/sptk.h>
#include <sptk3/CRWLock.h>
#include <time.h>
#include <errno.h>

using namespace sptk;

#define MAX_SEMAPHOR_COUNT  16384

#ifndef _WIN32
#if HAVE_PTHREAD_RWLOCK_TIMEDWRLOCK == 0
int pthread_rwlock_timedwrlock(pthread_rwlock_t * rwlock,
                                      const struct timespec * abs_timeout);
#endif
#if HAVE_PTHREAD_RWLOCK_TIMEDRDLOCK == 0
int pthread_rwlock_timedrdlock(pthread_rwlock_t * rwlock,
                                      const struct timespec * abs_timeout);
#endif
#endif

CRWLock::CRWLock() {
#ifndef _WIN32
    pthread_rwlock_init(&m_rwlock,NULL);
#else

    m_readLock = CreateEvent(0L,true,true,0L);
    SetEvent(m_readLock);
    m_writeLock = CreateMutex(0L,false,0L);
    InitializeCriticalSection(&m_criticalSection);
    m_readerCount = 0;
    m_writerMode = false;
#endif
}

CRWLock::~CRWLock() {
#ifndef _WIN32
    pthread_rwlock_destroy(&m_rwlock);
#else

    CloseHandle(m_readLock);
    CloseHandle(m_writeLock);
    DeleteCriticalSection(&m_criticalSection);
#endif
}

int CRWLock::lockR(int timeout) {
#ifndef _WIN32
    if (timeout > 0) {
        int   secs  = timeout / 1000;
        int   msecs = timeout % 1000;
        struct timespec   abstime = {
                                        time(NULL) + secs, msecs * 1000L
                                    };
        return pthread_rwlock_timedrdlock(&m_rwlock,&abstime);
    } else {
        return pthread_rwlock_rdlock(&m_rwlock) == 0;
    }
#else
    if (timeout < 0)
        timeout = INFINITE;

    // Trying to make sure we don't have an active writer
    if (WaitForSingleObject(m_writeLock, timeout) == WAIT_FAILED)
        return false;

    // Increment the readers counter.
    // Switch m_readLock to non-signaled state if needed,
    // blocking the potential waiters
    EnterCriticalSection(&m_criticalSection);
    if (m_readerCount == 0)
        ResetEvent(m_readLock);
    m_readerCount++;
    m_writerMode = false;
    LeaveCriticalSection(&m_criticalSection);
    return true;
#endif
}

int CRWLock::lockRW(int timeout) {
#ifndef _WIN32
    if (timeout > 0) {
        int   secs  = timeout / 1000;
        int   msecs = timeout % 1000;
        struct timespec   abstime = {
                                        time(NULL) + secs, msecs * 1000L
                                    };
        return pthread_rwlock_timedwrlock(&m_rwlock,&abstime) == 0;
    } else {
        return pthread_rwlock_wrlock(&m_rwlock) == 0;
    }
#else
    if (timeout < 0)
        timeout = INFINITE;

    // Trying to make sure we don't have active readers or writer
    // and to acquire the mutex
    HANDLE locks[] = { m_readLock, m_writeLock };
    if (WaitForMultipleObjects(2, locks, true, timeout) == WAIT_FAILED) {
        return false;
    }

    // Only one writer is possible
    m_readerCount = 1;
    m_writerMode = true;
    return true;
#endif
}

void CRWLock::unlock() {
#ifndef _WIN32
    pthread_rwlock_unlock(&m_rwlock);
#else
    // Decrement the readers counter.
    // Switch m_readLock to signaled state if needed, unblocking all the waiters
    EnterCriticalSection(&m_criticalSection);
    if (m_writerMode)
        ReleaseMutex(m_writeLock);
    if (m_readerCount == 1)
        SetEvent(m_readLock);
    m_readerCount--;
    LeaveCriticalSection(&m_criticalSection);
#endif
}

#ifndef _WIN32
#ifndef HAVE_PTHREAD_RWLOCK_TIMEDWRLOCK
/*!
    \fn static int pthread_rwlock_timedwrlock(pthread_rwlock_t *restrict rwlock, const struct timespec *restrict abs_timeout)
    This is ugly implemetation of timed rwlocking function, that has very low resolution, very high overhead
    (it's essantially is a spinlock with timeout) and generally should not have been written. But for poor OSs that don't
    have timers option implemented in pthreads (Solaris) it'll do it. I'm (iluxa) not using timeouts on RW locks anyways ;-)
 */
int pthread_rwlock_timedwrlock(pthread_rwlock_t * rwlock,
                                      const struct timespec * abs_timeout) {
    time_t tm,wt=abs_timeout->tv_sec;
    int rc;
    while((rc=pthread_rwlock_trywrlock(rwlock))==EBUSY) {
        time(&tm);
        if(tm>wt)
            return rc;
    }
    return rc;
}
#endif
#ifndef HAVE_PTHREAD_RWLOCK_TIMEDRDLOCK
int pthread_rwlock_timedrdlock(pthread_rwlock_t * rwlock,
                                      const struct timespec * abs_timeout) {
    time_t tm,wt=abs_timeout->tv_sec;
    int rc;
    while((rc=pthread_rwlock_tryrdlock(rwlock))==EBUSY) {
        time(&tm);
        if(tm>wt)
            return rc;
    }
    return rc;
}
#endif
#endif
